"""Contains the data storage classes for the arcos_gui.

The data storage classes are used to store the data and the settings for the
different widgets. Moste of the attributes contain callbacks functionallity
which are used to update the widgets when the data changes."""

from __future__ import annotations

from dataclasses import dataclass, field, fields, is_dataclass
from typing import Any, Callable, Generic, Literal, Sequence, TypeVar, Union, cast

import pandas as pd
import yaml
from arcos_gui.tools import ALLOWED_SETTINGS, ARCOSPARAMETERS_DEFAULTS


@dataclass
class value_changed:
    """Inner class to provide a more intuitive API for connecting and disconnecting callbacks."""

    _outer_instance: value_callback

    def connect(self, callback: Callable) -> None:
        """Register a callback function."""
        if callback not in self._outer_instance._callbacks and callable(callback):
            self._outer_instance._callbacks.append(callback)
        else:
            raise ValueError(
                "Callback is already registered or not callable. Use disconnect() to unregister."
            )

    def disconnect(self, callback: Callable) -> None:
        """Unregister a callback function."""
        if callback in self._outer_instance._callbacks:
            self._outer_instance._callbacks.remove(callback)
        else:
            raise ValueError("Callback is not registered.")


T = TypeVar("T")


@dataclass
class value_callback(Generic[T]):
    _value: T
    allowed_types: tuple[type, ...] = field(default_factory=tuple)
    value_changed: value_changed = field(init=False)
    _callbacks: list[Callable[[], None]] = field(default_factory=list, init=False)
    callbacks_blocked: bool = False
    value_name: str = "value_callback"
    verbose: bool = False
    missed_callbacks: int = 0

    def __post_init__(self):
        self.value_changed = value_changed(self)

    @property
    def value(self) -> T:
        return self._value

    @value.setter
    def value(self, value: T):
        if not isinstance(value, self.allowed_types):
            raise TypeError(
                f"Value for {self.value_name} must be of type {type(self._value)}"
            )

        self._value = cast(T, value)
        self._notify_observers()

    def _notify_observers(self):
        if self.callbacks_blocked:
            self.missed_callbacks += 1
            return
        self._emit()

    def _emit(self):
        for callback in self._callbacks:
            if self.verbose:
                print(
                    f"value_callback: {self.value_name} changed. Executing {callback}"
                )
            callback()

    def emit(self, only_missed: bool = False):
        if only_missed and self.missed_callbacks > 0:
            self._emit()
            self.missed_callbacks = 0
        else:
            self._notify_observers()

    def toggle_callback_block(self, set_explicit: bool | None = None):
        if set_explicit is not None:
            self.callbacks_blocked = set_explicit
        else:
            self.callbacks_blocked = not self.callbacks_blocked
        if not self.callbacks_blocked:
            self.missed_callbacks = 0

    def toggle_verbose(self):
        self.verbose = not self.verbose

    def __repr__(self) -> str:
        return repr(self._value)

    def __eq__(self, other: Any) -> bool:
        return self._value == other


@dataclass
class columnnames:
    """Stores column names for the dataframes.

    Attributes
    ----------
    frame_column: str
        The name of the column that contains the frame number.
    position_id: str
        The name of the column that contains the position id.
    object_id: str
        The name of the column that contains the object id.
    x_column: str
        The name of the column that contains the x coordinate.
    y_column: str
        The name of the column that contains the y coordinate.
    z_column: str
        The name of the column that contains the z coordinate.
    measurement_column_1: str
        The name of the column that contains the first measurement.
    measurement_column_2: str
        The name of the column that contains the second measurement.
    additional_filter_column: str
        The name of the column that contains the additional filter.
    measurement_math_operatoin: str
        The name of the column that contains the measurement math operation.
    measurement_bin: str | None
        The name of the column that contains the measurement bin.
    measurement_resc: str | None
        The name of the column that contains the measurement resc.
    collid_name: str
        The name of the column that contains the collid.
    measurement_column: str
        The name of the column that contains the measurement.
    pickablepickable_columns_names: list
        A list of all column names that can be set in the columnpicker_widget.
    posCol: list
        A list of all column names that are used for the position calculation.
        returns values for x, y and optionally z, if z is not 'None'.
    vcolscore: list
        A list of all column names that are used for the adding layers.
    as_dataframe: pandas.DataFrame
        A pandas.DataFrame containing all column names.
    """

    # working on replacing "None" comparisons with direct comparisons to None
    frame_column: str = "frame"
    position_id: Union[str, None] = "position_id"
    object_id: Union[str, None] = "track_id"
    x_column: str = "x"
    y_column: str = "y"
    z_column: Union[str, None] = "z"
    measurement_column_1: str = "measurement"
    measurement_column_2: Union[str, None] = "measurement_2"
    additional_filter_column: Union[str, None] = "additional_filter"
    measurement_math_operation: Union[str, None] = None
    measurement_bin: Union[str, None] = None
    measurement_resc: Union[str, None] = None
    collid_name: str = "collid"
    measurement_column: str = "measurement"

    @property
    def pickablepickable_columns_names(self):
        """returns a list of all column names that can be set in the columnpicker_widget"""
        return [
            self.frame_column,
            self.object_id,
            self.x_column,
            self.y_column,
            self.z_column,
            self.measurement_column_1,
            self.measurement_column_2,
            self.position_id,
            self.additional_filter_column,
            self.measurement_math_operation,
        ]

    @property
    def posCol(self):
        """returns a list of all coordinate columns"""
        if self.z_column:
            return [self.x_column, self.y_column, self.z_column]
        return [self.x_column, self.y_column]

    @property
    def vcolscore(self):
        """returns a list of core columns columns"""
        if self.z_column:
            return [self.frame_column, self.y_column, self.x_column, self.z_column]
        return [self.frame_column, self.y_column, self.x_column]

    @property
    def as_dataframe(self):
        """creates a dataframe from the columnnames"""
        data = [(attr.name, getattr(self, attr.name)) for attr in fields(self)]
        df = pd.DataFrame(columns=["Column", "value"], data=data)
        df["value"] = df["value"].astype(str)
        return df


def create_value_callback(default, name):
    return field(default_factory=lambda: value_callback(default[name], value_name=name))


@dataclass(frozen=True)
class ArcosParameters:
    """Stores the parameters for the arcos algorithm that can be set in the arcos widget.

    Attributes
    ----------
    interpolate_meas: bool
        wether to interpolate the measurements or not.
    clip_meas: bool
        wether to clip the measurements or not.
    clip_low: float
        the lower bound for the clipping.
    clip_high: float,
        the upper bound for the clipping.
    smooth_k: int,
        the smoothing kernel size.
    bias_k: int,
        the bias kernel size, for long term detrending.
    bias_method: str,
        the bias method.
    polyDeg: int,
        the polynomial degree, used for lm detrending.
    bin_threshold: float,
        the bin threshold, used for binarization.
    bin_peak_threshold: float,
        the bin peak threshold, used for binarization.
    eps_method: str,
        the eps method, used for estimating the eps parameter.
    neighbourhood_size: float,
        the eps parameter, used for the clustering within arcos.
    epsPrev: float,
        epsilon used for linking.
    min_cluster_size: int,
        the minimum cluster size, used for the clustering within arcos.
    nprev: int,
        the number of previous frames to consider for linking.
    min_dur: int,
        the minimum duration of a collective event.
    total_event_size: int,
        the minimum number of unique objects in a collective event.
    add_convex_hull: bool,
        wether to add the convex hull layer.
    as_dataframe: pandas.DataFrame
         a pandas.DataFrame containing all parameters.

    Methods
    -------
    set_all_parameters(parameters: arcos_parameters)
        set all parameters from another arcos_parameters object.
    reset_all_parameters()
        reset all parameters to their default values.
    set_verbose(verbose: bool)
        set the verbose parameter. If true, when a a callback is triggered,
        the callback function will be printed to the console.
    toggle_callback_block(block: bool)
        block or unblock all callbacks.
    """

    interpolate_meas: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            False, (bool,), value_name="interpolate_meas"
        )
    )
    clip_meas: value_callback[bool] = field(
        default_factory=lambda: value_callback(False, (bool,), value_name="clip_meas")
    )
    clip_low: value_callback[float] = field(
        default_factory=lambda: value_callback(0.0, (float,), value_name="clip_low")
    )
    clip_high: value_callback[float] = field(
        default_factory=lambda: value_callback(1.0, (float,), value_name="clip_high")
    )
    smooth_k: value_callback[int] = field(
        default_factory=lambda: value_callback(1, (int,), value_name="smooth_k")
    )
    bias_k: value_callback[int] = field(
        default_factory=lambda: value_callback(5, (int,), value_name="bias_k")
    )
    bias_method: value_callback[str] = field(
        default_factory=lambda: value_callback("none", (str,), value_name="bias_method")
    )
    polyDeg: value_callback[int] = field(
        default_factory=lambda: value_callback(1, (int,), value_name="polyDeg")
    )
    bin_threshold: value_callback[float] = field(
        default_factory=lambda: value_callback(
            0.5, (float, int), value_name="bin_threshold"
        )
    )
    bin_peak_threshold: value_callback[float] = field(
        default_factory=lambda: value_callback(
            0.5, (float, int), value_name="bin_peak_threshold"
        )
    )
    eps_method: value_callback[str] = field(
        default_factory=lambda: value_callback(
            "manual", (str,), value_name="eps_method"
        )
    )
    neighbourhood_size: value_callback[float] = field(
        default_factory=lambda: value_callback(
            20.0, (float, int), value_name="neighbourhood_size"
        )
    )
    epsPrev: value_callback[float] = field(
        default_factory=lambda: value_callback(
            20.0, (float, int, type(None)), value_name="epsPrev"
        )
    )
    min_clustersize: value_callback[int] = field(
        default_factory=lambda: value_callback(5, (int,), value_name="min_clustersize")
    )
    nprev: value_callback[int] = field(
        default_factory=lambda: value_callback(1, (int,), value_name="nprev")
    )
    min_dur: value_callback[int] = field(
        default_factory=lambda: value_callback(1, (int,), value_name="min_dur")
    )
    total_event_size: value_callback[int] = field(
        default_factory=lambda: value_callback(5, (int,), value_name="total_event_size")
    )
    add_convex_hull: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            True, (bool,), value_name="add_convex_hull"
        )
    )
    add_all_cells: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            True, (bool,), value_name="add_all_cells"
        )
    )
    add_bin_cells: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            True, (bool,), value_name="add_bin_cells"
        )
    )
    bin_advanded_settings: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            False, (bool,), value_name="bin_advanded_settings"
        )
    )
    detect_advanced_options: value_callback[bool] = field(
        default_factory=lambda: value_callback(
            False, (bool,), value_name="detect_advanced_options"
        )
    )

    @property
    def as_dataframe(self):
        """creates a dataframe from the arcos parameters"""
        data = [(attr.name, getattr(self, attr.name).value) for attr in fields(self)]
        df = pd.DataFrame(columns=["parameter", "value"], data=data)
        df["value"] = df["value"].astype(str)
        return df

    def set_all_parameters(
        self, arcos_parameters_object: ArcosParameters, trigger_callback: bool = True
    ):
        """sets all parameters to the values of another ArcosParameters object."""
        self.toggle_callback_block(not trigger_callback)
        for _field in fields(arcos_parameters_object):
            getattr(self, _field.name).value = getattr(
                arcos_parameters_object, _field.name
            ).value
        self.toggle_callback_block(False)

    def reset_all_parameters(self, trigger_callback: bool = True):
        """resets all values to their default values."""
        self.toggle_callback_block(not trigger_callback)
        for _field in fields(self):
            getattr(self, _field.name).value = ARCOSPARAMETERS_DEFAULTS[_field.name]
        self.toggle_callback_block(False)

    def set_verbose(self, verbose: bool = False):
        """sets the verbose parameter for all attributes."""
        for _field in fields(self):
            getattr(self, _field.name).verbose = verbose

    def toggle_callback_block(self, block: Union[bool, None] = None):
        """blocks or unblocks all callbacks."""
        for _field in fields(self):
            getattr(self, _field.name).toggle_callback_block(block)

    def emit(self, only_missed: bool = False):
        """emits all callbacks.

        Parameters
        ----------
        only_missed : bool, optional
            if True, only the missed callbacks will be emitted, by default False.
        """
        for _field in fields(self):
            getattr(self, _field.name).emit(only_missed=only_missed)


@dataclass(frozen=True)
class DataStorage:
    """Stores data for the GUI.

    Attributes are stored as a value_callback object.
    Both of these objects are wrappers around the actual data that are used to trigger the
    appropriate actions when the data is changed. Accessing the attributes value


    Attributes
    ----------
    file_name : file name of the input file
    original_data : original, unfiltered data
    filtered_data : filtered data based on the filter parameters set in the filter widget
    arcos_binarization : stores the binarization of the data for arcos
    arcos_output : stores the most recent output from arcos
    arcos_stats : stores the most recent stats from arcos
    columns : stores the column names of the data relevant for the analysis
    arcos_parameters : stores the parameters for arcos
    timestamp_parameters : stores the parameters for the timestamp dialog
    min_max_meas: stores the min and max values of the measurement

    Methods
    -------
    reset_all_attributes(trigger_callbacks=False)
        Resets all attributes to their default values. If trigger_callbacks is True, the
        callbacks of all attributes are triggered.
    reset_relevant_attributes(trigger_callbacks=False)
        Resets all attributes that are relevant for the analysis to their default values.
        If trigger_callbacks is True, the callbacks of all attributes are triggered.
    """

    file_name: value_callback[str] = field(
        default_factory=lambda: value_callback(".", (str,), value_name="file_name")
    )
    original_data: value_callback[pd.DataFrame] = field(
        default_factory=lambda: value_callback(
            pd.DataFrame(), (pd.DataFrame,), value_name="original_data"
        )
    )
    filtered_data: value_callback[pd.DataFrame] = field(
        default_factory=lambda: value_callback(
            pd.DataFrame(), (pd.DataFrame,), value_name="filtered_data"
        )
    )
    arcos_binarization: value_callback[pd.DataFrame] = field(
        default_factory=lambda: value_callback(
            pd.DataFrame(), (pd.DataFrame,), value_name="arcos_binarization"
        )
    )
    arcos_output: value_callback[pd.DataFrame] = field(
        default_factory=lambda: value_callback(
            pd.DataFrame(), (pd.DataFrame,), value_name="arcos_output"
        )
    )
    arcos_stats: value_callback[pd.DataFrame] = field(
        default_factory=lambda: value_callback(
            pd.DataFrame(), (pd.DataFrame,), value_name="arcos_stats"
        )
    )
    columns: value_callback[columnnames] = field(
        default_factory=lambda: value_callback(
            columnnames(), (columnnames,), value_name="columns"
        )
    )
    arcos_parameters: value_callback[ArcosParameters] = field(
        default_factory=lambda: value_callback(
            ArcosParameters(), (ArcosParameters,), value_name="arcos_parameters"
        )
    )
    min_max_meas: value_callback[list] = field(
        default_factory=lambda: value_callback(
            [0.0, 0.5], (list,), value_name="min_max_meas"
        )
    )
    point_size: value_callback[float] = field(
        default_factory=lambda: value_callback(
            10.0, (float, int), value_name="point_size"
        )
    )
    selected_object_id: value_callback[int | None] = field(
        default_factory=lambda: value_callback(
            None, (list, int, type(None)), value_name="selected_object_id"
        )
    )
    lut: value_callback[str] = field(
        default_factory=lambda: value_callback("inferno", (str,), value_name="lut")
    )
    output_order: value_callback[str] = field(
        default_factory=lambda: value_callback(
            "txyz", (str,), value_name="output_order"
        )
    )
    min_max_tracklenght: value_callback[list] = field(
        default_factory=lambda: value_callback(
            [0, 1], (list,), value_name="min_max_tracklenght"
        )
    )

    def export_to_yaml(self, filepath: str):
        """Exports all attributes of DataStorage class to a YAML file, excluding those which are DataFrames.

        Parameters:
        filepath (str): The file path where the YAML file should be saved.
        """

        def extract_values(obj):
            """Recursive function to extract values from dataclasses and value_callback objects."""
            if isinstance(obj, value_callback):
                obj = obj.value
            if is_dataclass(obj):
                return {
                    f.name: extract_values(getattr(obj, f.name)) for f in fields(obj)
                }
            else:
                return obj

        data_dict = {
            f.name: extract_values(getattr(self, f.name))
            for f in fields(self)
            if not isinstance(getattr(self, f.name).value, pd.DataFrame)
        }

        # Writing the data dictionary to a YAML file
        with open(filepath, "w") as file:
            yaml.safe_dump(data_dict, file)

    def import_from_yaml(
        self,
        filepath: str,
        selected_attributes: (
            Sequence[
                Literal[
                    "file_name",
                    "columns",
                    "arcos_parameters",
                    "min_max_meas",
                    "point_size",
                    "lut",
                    "output_order",
                    "min_max_tracklenght",
                ]
            ]
            | None
        ) = None,
    ):
        """Imports the parameters from a YAML file and updates the attributes of DataStorage class.

        Parameters:
        filepath (str): The file path where the YAML file is located.
        selected_attributes (Sequence[allowed_import_values] | None): A list of attributes to be imported.
            If None, all attributes will be imported. Defaults to None.
            Note: importing columns will reset all attributes.
        """
        if selected_attributes is None:
            _selected_attributes = cast(
                Sequence[
                    Literal[
                        "file_name",
                        "columns",
                        "arcos_parameters",
                        "min_max_meas",
                        "point_size",
                        "lut",
                        "output_order",
                        "min_max_tracklenght",
                    ]
                ],
                ALLOWED_SETTINGS,
            )
        else:
            _selected_attributes = selected_attributes

        # check if columns or file_name are in _selected_attributes
        if any(
            attr in _selected_attributes
            for attr in ["columns", "file_name", "min_max_tracklenght"]
        ):
            self.reset_all_attributes()
            print("resetting all attributes")

        for attr in _selected_attributes:
            if attr not in ALLOWED_SETTINGS:
                raise ValueError(f"Cant import {attr} from YAML file.")

        self.toggle_callback_block(
            True
        )  # black all callbacks while importing to avoid unnecessary updates

        def update_attributes(obj, data_dict, skip_selected_check=True):
            """Recursive function to update attributes from a nested dictionary."""
            for key, value in data_dict.items():
                if key not in _selected_attributes and skip_selected_check:
                    continue
                attr = getattr(obj, key)
                if isinstance(attr, value_callback):
                    if is_dataclass(attr.value):
                        update_attributes(attr.value, value, skip_selected_check=False)
                    else:
                        attr.value = value
                else:
                    setattr(obj, key, value)

        with open(filepath) as file:
            data_dict = yaml.safe_load(file)

        update_attributes(self, data_dict)

        # manually trigger callbacks for all attributes that were not updated
        self.emit_callbacks(only_missed=True)
        self.toggle_callback_block(False)

    def reset_all_attributes(self, trigger_callback=False):
        """resets all attributes to their default values.

        Parameters
        ----------
        trigger_callback : bool, optional
            if True, the callback function of the value_callback attributes will be triggered, by default False.
        """
        self.toggle_callback_block(not trigger_callback)
        self.file_name.value = "."
        self.original_data.value = pd.DataFrame()
        self.filtered_data.value = pd.DataFrame()
        self.arcos_binarization.value = pd.DataFrame()
        self.arcos_output.value = pd.DataFrame()
        self.arcos_stats.value = pd.DataFrame()
        self.columns.value = columnnames()
        self.arcos_parameters.value.reset_all_parameters()
        self.min_max_meas.value = [0, 0.5]
        self.point_size.value = 10
        self.selected_object_id.value = None
        self.lut.value = "inferno"
        self.output_order.value = "txyz"
        self.min_max_tracklenght.value = [0, 1]
        self.toggle_callback_block(False)

    def reset_relevant_attributes(self, trigger_callback=False):
        """Resets relevant attributes to their default values.

        Parameters
        ----------
        trigger_callback : bool, optional
            If True, the callback function of the attributes will be triggered, by default False.
        """
        self.toggle_callback_block(not trigger_callback)
        self.filtered_data.value = pd.DataFrame()
        self.arcos_binarization.value = pd.DataFrame()
        self.arcos_output.value = pd.DataFrame()
        self.arcos_stats.value = pd.DataFrame()
        self.selected_object_id.value = None
        self.toggle_callback_block(False)

    def reset_arcos_data(self):
        """Resets all arcos related attributes to their default values."""
        self.arcos_binarization.value = pd.DataFrame()
        self.arcos_output.value = pd.DataFrame()
        self.arcos_stats.value = pd.DataFrame()

    def toggle_callback_block(self, block: bool | None):
        def recursive_toggle(obj, block):
            for _field in fields(obj):
                value_callback_field = getattr(obj, _field.name)
                if isinstance(value_callback_field, value_callback):
                    value_callback_field.toggle_callback_block(block)
                    if hasattr(value_callback_field, "value"):
                        nested_value = value_callback_field.value
                        if is_dataclass(nested_value):
                            recursive_toggle(nested_value, block)

        recursive_toggle(self, block)

    def set_callbacks_verbose(self, verbose: bool):
        def recursive_set_verbose(obj, verbose):
            for _field in fields(obj):
                value_callback_field = getattr(obj, _field.name)
                if isinstance(value_callback_field, value_callback):
                    value_callback_field.verbose = verbose
                    if hasattr(value_callback_field, "value"):
                        nested_value = value_callback_field.value
                        if is_dataclass(nested_value):
                            recursive_set_verbose(nested_value, verbose)

        recursive_set_verbose(self, verbose)

    def emit_callbacks(self, only_missed: bool = False):
        def recursive_emit(obj, only_missed):
            for _field in fields(obj):
                value_callback_field = getattr(obj, _field.name)
                if isinstance(value_callback_field, value_callback):
                    value_callback_field.emit(only_missed=only_missed)
                    if hasattr(value_callback_field, "value"):
                        nested_value = value_callback_field.value
                        if is_dataclass(nested_value):
                            recursive_emit(nested_value, only_missed)

        recursive_emit(self, only_missed)

    def load_data(self, filename, trigger_callback=True):
        """Loads data from a csv file."""
        if trigger_callback:
            self.original_data.value = pd.read_csv(filename)
        else:
            self.original_data._value = pd.read_csv(filename)

    def __eq__(self, other):
        if not isinstance(other, DataStorage):
            return False

        dataframes_to_check = [
            ("original_data", self.original_data.value, other.original_data.value),
            ("filtered_data", self.filtered_data.value, other.filtered_data.value),
            (
                "arcos_binarization",
                self.arcos_binarization.value,
                other.arcos_binarization.value,
            ),
            ("arcos_output", self.arcos_output.value, other.arcos_output.value),
            ("arcos_stats", self.arcos_stats.value, other.arcos_stats.value),
        ]

        for name, df1, df2 in dataframes_to_check:
            if not self._are_dataframes_equal(df1, df2):
                return False

        attributes_to_check = [
            "columns",
            "file_name",
            "arcos_parameters",
            "min_max_meas",
            "point_size",
            "selected_object_id",
            "lut",
        ]

        for attr in attributes_to_check:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def _are_dataframes_equal(self, df1, df2):
        try:
            pd.testing.assert_frame_equal(df1, df2)
            return True
        except AssertionError:
            return False



"""Widgets for plotting time series data and collective events."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import napari
from arcos_gui.tools._plots import CollevPlotter, NoodlePlot, TimeSeriesPlots
from qtpy import QtCore, QtWidgets
from qtpy.QtGui import QIcon

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage


class DataPlot(QtWidgets.QDialog):
    """Popup Dialog for plotting time series data."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.tsplot_layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.tsplot_layout)


# icons
ICONS = Path(__file__).parent.parent / "_icons"

# still need to rewrite actual plots to be a bit nicer


class tsPlotWidget(QtWidgets.QWidget):
    """Widget for plotting time series data."""

    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        data_storage_instance: DataStorage,
        parent=None,
    ):
        super().__init__(parent)
        self.viewer = viewer
        self._data_storage_instance = data_storage_instance
        self.timeseriesplot = TimeSeriesPlots(parent=self)
        self.plot_dialog_data = DataPlot(parent=self)
        self._add_plot_widgets()
        self._add_icon()
        self._data_storage_instance.original_data.value_changed.connect(
            self._on_data_update
        )
        self._data_storage_instance.arcos_binarization.value_changed.connect(
            self._on_data_update
        )
        self._data_storage_instance.selected_object_id.value_changed.connect(
            self._on_data_update
        )

    def _on_data_update(self):
        self._on_data_clear()
        df_orig = self._data_storage_instance.original_data.value
        df_bin = self._data_storage_instance.arcos_binarization.value
        frame_col = self._data_storage_instance.columns.value.frame_column
        object_id_col = self._data_storage_instance.columns.value.object_id
        measurement_col = self._data_storage_instance.columns.value.measurement_column
        x_coord_col = self._data_storage_instance.columns.value.x_column
        y_coord_col = self._data_storage_instance.columns.value.y_column
        measurement_resc_col = (
            self._data_storage_instance.columns.value.measurement_resc
        )
        object_id_number = self._data_storage_instance.selected_object_id.value
        self.timeseriesplot.update_plot(
            df_orig,
            df_bin,
            frame_col,
            object_id_col,
            x_coord_col,
            y_coord_col,
            measurement_col,
            measurement_resc_col,
            object_id_number,
        )

    def _on_data_clear(self):
        self.timeseriesplot.data_clear()

    def _add_icon(self):
        expand_plot_icon = QIcon(str(ICONS / "enlarge_window.png"))
        self.expand_plot.setIcon(expand_plot_icon)
        self.expand_plot.setMaximumSize(30, 30)

    def _open_data_plot(self):
        self.plot_dialog_data.closeEvent = self._close_plot
        self.plot_dialog_data.tsplot_layout.addWidget(self.timeseriesplot)
        self.plot_dialog_data.show()
        self.timeseriesplot_groupbox.hide()

    def _close_plot(self, event):
        self.timeseriesplot_groupbox.layout().removeWidget(self.expand_plot)
        self.tsplot_layout.addWidget(self.timeseriesplot, alignment=QtCore.Qt.AlignTop)
        self.timeseriesplot_groupbox.layout().addWidget(
            self.expand_plot, alignment=QtCore.Qt.AlignRight
        )
        self.timeseriesplot_groupbox.show()

    def _add_plot_widgets(self):
        self.timeseriesplot_groupbox = QtWidgets.QGroupBox("Time Series Plot")
        self.widget_layout = QtWidgets.QVBoxLayout()
        self.tsplot_layout = QtWidgets.QVBoxLayout()
        self.expand_plot = QtWidgets.QPushButton()
        self.timeseriesplot_groupbox.setLayout(self.tsplot_layout)
        self.setLayout(self.widget_layout)
        self.tsplot_layout.addWidget(self.timeseriesplot)
        self.widget_layout.addWidget(self.timeseriesplot_groupbox)
        self.timeseriesplot_groupbox.layout().addWidget(
            self.expand_plot, alignment=QtCore.Qt.AlignRight
        )
        self.expand_plot.clicked.connect(self._open_data_plot)


class collevPlotWidget(QtWidgets.QWidget):
    """Widget for collective events plot."""

    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        data_storage_instance: DataStorage,
        parent=None,
    ):
        super().__init__(parent)
        self.viewer = viewer
        self._data_storage_instance = data_storage_instance
        self.collevplot = CollevPlotter(self.viewer, parent=self)
        self.noodle_plot = NoodlePlot(self.viewer, parent=self)
        self.tab_widget = QtWidgets.QTabWidget()
        self.tab_widget.addTab(self.collevplot, "Collev Plot")
        self.tab_widget.addTab(self.noodle_plot, "Noodle Plot")
        self.plot_dialog_collev = DataPlot(parent=self)
        self._add_plot_widgets()
        self._add_icon()

        self._data_storage_instance.arcos_binarization.value_changed.connect(
            self._data_clear
        )
        self._data_storage_instance.arcos_output.value_changed.connect(
            self._on_data_update
        )
        self._data_storage_instance.original_data.value_changed.connect(
            self._data_clear
        )
        self.tab_widget.setCurrentIndex(1)

    def _on_data_update(self):
        frame_col = self._data_storage_instance.columns.value.frame_column
        oid_col = self._data_storage_instance.columns.value.object_id
        x_coord = self._data_storage_instance.columns.value.x_column
        y_coord = self._data_storage_instance.columns.value.y_column
        z_coord = self._data_storage_instance.columns.value.z_column
        point_size = self._data_storage_instance.point_size.value
        arcos_data = self._data_storage_instance.arcos_output.value
        output_order = self._data_storage_instance.output_order.value
        arcos_stats = self._data_storage_instance.arcos_stats.value

        self.collevplot.update_plot(
            frame_col,
            oid_col,
            x_coord,
            y_coord,
            z_coord,
            arcos_data,
            arcos_stats,
            point_size,
            output_order,
        )

        self.noodle_plot.update_plot(
            frame_col,
            oid_col,
            x_coord,
            y_coord,
            z_coord,
            arcos_data,
            point_size,
            output_order,
        )

    def _data_clear(self):
        self.collevplot.clear_plot()
        self.noodle_plot.clear_plot()

    def _open_plot(self):
        self.plot_dialog_collev.closeEvent = self._close_plot
        self.plot_dialog_collev.tsplot_layout.addWidget(self.tab_widget)
        self.plot_dialog_collev.show()
        self.collevplot_goupbox.hide()

    def _close_plot(self, event):
        self.collevplot_goupbox.layout().removeWidget(self.expand_plot)
        self.evplot_layout.addWidget(self.tab_widget, alignment=QtCore.Qt.AlignTop)
        self.collevplot_goupbox.layout().addWidget(
            self.expand_plot, alignment=QtCore.Qt.AlignRight
        )
        self.collevplot_goupbox.show()

    def _add_plot_widgets(self):
        self.collevplot_goupbox = QtWidgets.QGroupBox("Event Plot")
        self.widget_layout = QtWidgets.QVBoxLayout()
        self.evplot_layout = QtWidgets.QVBoxLayout()
        self.expand_plot = QtWidgets.QPushButton()
        self.collevplot_goupbox.setLayout(self.evplot_layout)
        self.widget_layout.addWidget(self.collevplot_goupbox)
        self.collevplot_goupbox.layout().addWidget(self.tab_widget)
        self.collevplot_goupbox.layout().addWidget(
            self.expand_plot, alignment=QtCore.Qt.AlignRight
        )
        self.expand_plot.clicked.connect(self._open_plot)
        self.setLayout(self.widget_layout)

    def _add_icon(self):
        expand_plot_icon = QIcon(str(ICONS / "enlarge_window.png"))
        self.expand_plot.setIcon(expand_plot_icon)
        self.expand_plot.setMaximumSize(30, 30)


if __name__ == "__main__":
    import sys

    from arcos_gui.processing import DataStorage  # noqa: F811

    viewer = napari.Viewer()
    app = QtWidgets.QApplication(sys.argv)
    ds = DataStorage()
    window = collevPlotWidget(viewer, ds)
    window.show()
    sys.exit(app.exec_())



"""Module contiaing serveral utility functions to prepare layer data tuples."""

from __future__ import annotations

from typing import Union

import numpy as np
import pandas as pd
from arcos_gui.tools import (
    ARCOS_LAYERS,
    COLOR_CYCLE,
    TAB20,
    fix_3d_convex_hull,
    get_verticesHull,
    make_surface_3d,
    reshape_by_input_string,
)


def prepare_all_cells_layer(
    df_all: pd.DataFrame,
    vColsCore: list[str | None],
    track_id_col: str | None,
    measurement_name: str,
    lut: str,
    min_max: list[float],
    size: float,
    axis_order: str | None = None,
) -> Union[tuple, None]:
    """Prepare all cells layer.

    Parameters
    ----------
    df_all : pd.DataFrame
        dataframe with all cells
    vColsCore : list
        list with core columns in dataframe
        order: [framecol, ycol, xcol, zcol(optional)]
    measurement_name : str
        name of measurement
    lut : str
        name of lut
    min_max : tuple
        min and max values for measurement
    size : float
        size of cells
    axis_order : str
        order of axis, e.g. 'tzyx', possible values: ['t', 'z', 'y', 'x'].
        Default: 'tzyx' for 3D data, 'tyx' for 2D data
    Returns
    -------
    all_cells_layer : dict
        dictionary with all cells layer
    """
    # np matrix with all cells
    df_all = df_all.copy()
    df_all.interpolate(method="linear", inplace=True)
    data_all_np = df_all[vColsCore].to_numpy()
    data_all_np = reshape_by_input_string(data_all_np, axis_order, vColsCore)

    if track_id_col:
        data_id_np = df_all[track_id_col].to_numpy()
    else:
        data_id_np = np.zeros(data_all_np.shape[0])

    # a dictionary with activities;
    # shown as a color code of all cells
    data_all_prop = {"act": df_all[measurement_name].astype(float), "id": data_id_np}
    # tuple to return layer as layer.data.tuple
    all_cells = (
        data_all_np,
        {
            "properties": data_all_prop,
            "edge_width": 0,
            "edge_color": "act",
            "face_color": "act",
            "face_colormap": lut,
            "face_contrast_limits": tuple(min_max),
            "size": size,
            "opacity": 1,
            "symbol": "disc",
            "name": ARCOS_LAYERS["all_cells"],
        },
        "points",
    )
    return all_cells


def prepare_active_cells_layer(
    df_bin: pd.DataFrame,
    vColsCore: list,
    measbin_col: str | None,
    size: float,
    axis_order: str | None = None,
    padd_time: bool = True,
) -> Union[tuple, None]:
    """Prepare active cells layer.

    Parameters
    ----------
    df_bin : pd.DataFrame
        dataframe with binarized data
    vColsCore : list
        list with core columns in dataframe
        order: [framecol, ycol, xcol, zcol(optional)]
    measbin_col : str
        name of binarized measurement column
    size : float
        size of cells
    axis_order : str
        order of axis, e.g. 'tzyx', possible values: ['t', 'z', 'y', 'x'].
        Default: 'tzyx' for 3D data, 'tyx' for 2D data.
    padd_time : bool
        if True, add a row with timepoint 0 if not present
        to make sure the timeaxis starts with 0
    Returns
    -------
    active_cells_layer : tuple
        tuple with active cells layer
    """

    # np matrix with acvtive cells; shown as black dots
    df_bin_filtered = df_bin[df_bin[measbin_col] > 0]
    datAct = df_bin_filtered[vColsCore].to_numpy()
    datAct = reshape_by_input_string(datAct, axis_order, vColsCore)

    # check if datAct starts with timepoint 0 if not add a row with timepoint 0
    if datAct[0][0] != 0 and padd_time:
        # take first timepoint and replace it with 0
        dat_tp0 = datAct[0].copy()
        dat_tp0[0] = 0
        datAct = np.insert(datAct, 0, dat_tp0, axis=0)
        shown_points = np.repeat(True, datAct.shape[0])
        shown_points[0] = False
    else:
        shown_points = np.repeat(True, datAct.shape[0])

    if datAct.size == 0:
        return None

    active_cells = (
        datAct,
        {
            "size": round(size / 2.5, 2),
            "edge_width": 0,
            "face_color": "black",
            "opacity": 1,
            "symbol": "disc",
            "name": ARCOS_LAYERS["active_cells"],
            "shown": shown_points,
        },
        "points",
    )

    return active_cells


def prepare_events_layer(
    df_coll: pd.DataFrame,
    vColsCore: list,
    size: float,
    axis_order: str | None = None,
    padd_time: bool = True,
) -> Union[tuple, None]:
    """Prepare events layer.

    Parameters
    ----------
    df_coll : pd.DataFrame
        dataframe with events
    vColsCore : list
        list with core columns in dataframe
        order: [framecol, ycol, xcol, zcol(optional)]
    size : float
        size of cells
    axis_order : str
        order of axis, e.g. 'tzyx', possible values: ['t', 'z', 'y', 'x'].
        Default: 'tzyx' for 3D data, 'tyx' for 2D data.
    padd_time : bool
        if True, add a row with timepoint 0 if not present
        to make sure the timeaxis starts with 0
    Returns
    -------
    coll_cells : tuple
        tuple with events layer
    """
    # np matrix with cells in collective events
    data_collevent_np = df_coll[vColsCore].to_numpy()
    data_collevent_np = reshape_by_input_string(
        data_collevent_np, axis_order, vColsCore
    )

    # create remaining layer.data.tuples
    np_clids = df_coll["collid"].to_numpy()

    if np_clids.size == 0:
        return None

    # check if datAct starts with timepoint 0 if not add a row with timepoint 0
    if data_collevent_np[0][0] != 0 and padd_time:
        dat_collev_0 = data_collevent_np[0].copy()
        dat_collev_0[0] = 0
        data_collevent_np = np.insert(data_collevent_np, 0, dat_collev_0, axis=0)
        np_clids = np.insert(np_clids, 0, 0, axis=0)
        shown_points = np.repeat(True, data_collevent_np.shape[0])
        shown_points[0] = False
    else:
        shown_points = np.repeat(True, data_collevent_np.shape[0])

    color_ids = np.take(np.array(COLOR_CYCLE), list(np_clids), mode="wrap")
    coll_cells = (
        data_collevent_np,
        {
            "face_color": color_ids,
            "size": round(size / 1.2, 2),
            "edge_width": 0,
            "opacity": 1,
            "name": ARCOS_LAYERS["collective_events_cells"],
            "shown": shown_points,
        },
        "points",
    )

    return coll_cells


def prepare_convex_hull_layer(
    df_filtered: pd.DataFrame,
    df_coll: pd.DataFrame,
    collid_name: str,
    vColsCore: list,
    axis_order: str | None = None,
) -> Union[tuple, None]:
    """Prepare convex hull layer.

    Parameters
    ----------
    df_filtered : pd.DataFrame
        dataframe with filtered data
    df_coll : pd.DataFrame
        dataframe with events
    collid_name : str
        name of collid column
    vColsCore : list
        list with core columns in dataframe
        order: [framecol, ycol, xcol, zcol(optional)]
    axis_order : str
        order of axis, e.g. 'tzyx', possible values: ['t', 'z', 'y', 'x'].
        Default: 'tzyx' for 3D data, 'tyx' for 2D data.

    Returns
    -------
    coll_hull : tuple
        tuple with convex hull layer
    """

    if df_coll[collid_name].to_numpy().size == 0:
        return None

    if len(vColsCore) == 3:
        # 2D data
        datChull, color_ids = get_verticesHull(
            df_coll,
            frame=vColsCore[0],
            colid=collid_name,
            col_x=vColsCore[2],
            col_y=vColsCore[1],
        )

        # order according to input string
        datChull = [
            reshape_by_input_string(i, input_string=axis_order, vColsCore=vColsCore)
            for i in datChull
        ]

        coll_events = (
            datChull,
            {
                "face_color": color_ids,
                "shape_type": "polygon",
                "text": None,
                "opacity": 0.5,
                "edge_color": "white",
                "edge_width": 0,
                "name": ARCOS_LAYERS["event_hulls"],
            },
            "shapes",
        )
        return coll_events

    # 3D data
    # reorder columns list to match axis order

    event_surfaces = make_surface_3d(
        df_coll,
        vColsCore=vColsCore,
        colid=collid_name,
        output_order=axis_order,
    )

    event_surfaces = fix_3d_convex_hull(
        df_filtered[vColsCore],
        event_surfaces[0],
        event_surfaces[1],
        event_surfaces[2],
        col_t=vColsCore[0],
    )

    coll_events = (
        event_surfaces,
        {
            "colormap": TAB20,
            "name": ARCOS_LAYERS["event_hulls"],
            "opacity": 0.5,
        },
        "surface",
    )
    return coll_events



"""Main module for napari plugin."""

import napari
from arcos_gui import open_plugin


def main():
    """Main function. Opens napari viewer and adds plugin dock widget."""
    viewer = napari.Viewer()
    open_plugin(viewer)
    napari.run()


if __name__ == "__main__":
    main()



"""Arcos GUI Plugin for Napari."""

__version__ = "0.1.3"


from arcos_gui._helper_functions import (
    filter_data,
    get_arcos_output,
    get_current_arcos_plugin,
    load_dataframe,
    load_dataframe_with_columnpicker,
    load_sample_data,
    open_plugin,
    run_arcos,
    run_binarization_only,
)
from napari.utils.notifications import show_info

show_info("Loading Plugin")

__all__ = [
    "open_plugin",
    "load_dataframe",
    "load_sample_data",
    "load_dataframe_with_columnpicker",
    "filter_data",
    "run_binarization_only",
    "run_arcos",
    "get_arcos_output",
    "get_current_arcos_plugin",
]



"""Module containing utility classes to make layers for napari."""

from __future__ import annotations

from typing import TYPE_CHECKING

from arcos_gui.tools import ARCOS_LAYERS
from napari.layers import Layer

from ._layer_data_tuple import (
    prepare_active_cells_layer,
    prepare_all_cells_layer,
    prepare_convex_hull_layer,
    prepare_events_layer,
)

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage
    from napari import viewer


class Layermaker:
    """Class to make layers for napari."""

    def __init__(self, viewer: viewer.Viewer, data_storage_instance: DataStorage):
        self.data_storage_instance = data_storage_instance
        self.viewer = viewer
        self.padd_time = True

    def _remove_old_layers(self):
        layers_names = [layer.name for layer in self.viewer.layers]
        for layer in ARCOS_LAYERS.values():
            if layer in layers_names:
                self.viewer.layers.remove(layer)

    def make_layers_bin(
        self, all_cells: bool | None = None, active_cells: bool | None = None
    ):
        """adds layers from self.layers_to_create,
        whitch itself is upated from run_arcos method"""
        if all_cells is None:
            all_cells = (
                self.data_storage_instance.arcos_parameters.value.add_all_cells.value
            )
        if active_cells is None:
            active_cells = (
                self.data_storage_instance.arcos_parameters.value.add_bin_cells.value
            )

        self._remove_old_layers()
        for result in self._layers_to_create_bin(
            all_cells=all_cells, active_cells=active_cells
        ):
            if result:
                self.viewer.add_layer(Layer.create(*result))
                if result[1]["name"] == ARCOS_LAYERS["all_cells"]:
                    self._connect_all_cells_point_select()

    def make_layers_all(
        self,
        convex_hull: bool | None = None,
        all_cells: bool | None = None,
        active_cells: bool | None = None,
    ):
        """adds layers from self.layers_to_create,
        whitch itself is upated from run_arcos method"""
        self._remove_old_layers()
        if convex_hull is None:
            convex_hull = (
                self.data_storage_instance.arcos_parameters.value.add_convex_hull.value
            )
        if all_cells is None:
            all_cells = (
                self.data_storage_instance.arcos_parameters.value.add_all_cells.value
            )
        if active_cells is None:
            active_cells = (
                self.data_storage_instance.arcos_parameters.value.add_bin_cells.value
            )

        for result in self._layers_to_create_all(convex_hull, all_cells, active_cells):
            if result:
                self.viewer.add_layer(Layer.create(*result))
                if result[1]["name"] == ARCOS_LAYERS["all_cells"]:
                    self._connect_all_cells_point_select()

    def _layers_to_create_all(
        self,
        convex_hull: bool = True,
        all_cells: bool = True,
        active_cells: bool = True,
    ):
        VcolsCore = self.data_storage_instance.columns.value.vcolscore
        layers: list = []
        layers = self._make_layers_bin(layers, VcolsCore, all_cells, active_cells)
        layers = self._make_layers_arcos(layers, VcolsCore, convex_hull)
        if len(VcolsCore) <= 3:
            self.viewer.dims.ndisplay = 2
        else:
            self.viewer.dims.ndisplay = 3
        return layers

    def _layers_to_create_bin(self, all_cells: bool = True, active_cells: bool = True):
        VcolsCore = self.data_storage_instance.columns.value.vcolscore
        layers: list = []
        layers = self._make_layers_bin(layers, VcolsCore, all_cells, active_cells)
        if len(VcolsCore) <= 3:
            self.viewer.dims.ndisplay = 2
        else:
            self.viewer.dims.ndisplay = 3
        return layers

    def _make_layers_bin(
        self,
        layers: list,
        VcolsCore: list,
        all_cells: bool = True,
        active_cells: bool = True,
    ):
        """returns a list of layers to be created"""
        VcolsCore = self.data_storage_instance.columns.value.vcolscore
        if self.data_storage_instance.arcos_binarization.value.empty:
            return layers
        if all_cells:
            layers.append(
                prepare_all_cells_layer(
                    self.data_storage_instance.filtered_data.value,
                    VcolsCore,
                    self.data_storage_instance.columns.value.object_id,
                    self.data_storage_instance.columns.value.measurement_column,
                    self.data_storage_instance.lut.value,
                    self.data_storage_instance.min_max_meas.value,
                    self.data_storage_instance.point_size.value,
                    self.data_storage_instance.output_order.value,
                )
            )
        if active_cells:
            layers.append(
                prepare_active_cells_layer(
                    self.data_storage_instance.arcos_binarization.value,
                    VcolsCore,
                    self.data_storage_instance.columns.value.measurement_bin,
                    self.data_storage_instance.point_size.value,
                    self.data_storage_instance.output_order.value,
                    self.padd_time,
                )
            )
        return layers

    def _make_layers_arcos(
        self, layers: list, VcolsCore: list, convex_hull: bool = True
    ):
        if self.data_storage_instance.arcos_output.value.empty:
            return layers
        layers.append(
            prepare_events_layer(
                self.data_storage_instance.arcos_output.value,
                VcolsCore,
                self.data_storage_instance.point_size.value,
                self.data_storage_instance.output_order.value,
                self.padd_time,
            )
        )

        if convex_hull:
            layers.append(
                prepare_convex_hull_layer(
                    self.data_storage_instance.filtered_data.value,
                    self.data_storage_instance.arcos_output.value,
                    self.data_storage_instance.columns.value.collid_name,
                    VcolsCore,
                    self.data_storage_instance.output_order.value,
                )
            )

        return layers

    def _connect_all_cells_point_select(self):
        self.viewer.layers["All Cells"].events.current_properties.connect(
            self._pick_event
        )
        self._prev_point_data = None
        self.point_index = None

    def _pick_event(self, event):
        if self._prev_point_data is not None:
            event.sources[0].data[self.point_index] = self._prev_point_data
            self.viewer.layers["All Cells"].refresh()

        self.point_index = list(event.sources[0].selected_data)
        self.id = event.sources[0].properties["id"][self.point_index]
        self._prev_point_data = event.sources[0].data[self.point_index]
        self.data_storage_instance.selected_object_id.value = self.id.tolist()



"""Module containing the bottom bar widget class."""

from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy import QtWidgets
from qtpy.QtCore import QUrl
from qtpy.QtGui import QDesktopServices

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage


class _bottombar_widget(QtWidgets.QWidget):
    """Bottom bar widget. Displays the number of detected collective events."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setupui()

    def setupui(self):
        """Setup the UI. Add widgets to the layout."""
        self.bottom_bar_layout = QtWidgets.QHBoxLayout()
        self.collev_number_display_label = QtWidgets.QLabel(
            "Number of detected collective events: "
        )
        self.collev_number_display = QtWidgets.QLCDNumber()
        self.collev_number_display.setDigitCount(4)
        self.collev_number_display.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.arcos_help_button = QtWidgets.QPushButton("Open Documentation")
        self.bottom_bar_layout.addWidget(self.collev_number_display_label)
        self.bottom_bar_layout.addWidget(self.collev_number_display)
        self.bottom_bar_layout.addWidget(self.arcos_help_button)
        self.setLayout(self.bottom_bar_layout)
        self._connect_signals()

    def _update_help_pressed(self):
        url = QUrl("https://bgraedel.github.io/arcos-gui/Usage/")
        QDesktopServices.openUrl(url)

    def _connect_signals(self):
        self.arcos_help_button.clicked.connect(self._update_help_pressed)


class BottombarController:
    """Controller for the bottom bar widget."""

    def __init__(self, data_storage_instance: DataStorage, parent=None):
        self.widget = _bottombar_widget(parent)
        self.data_storage_instance = data_storage_instance
        self._connect_signals()

    def _connect_signals(self):
        self.data_storage_instance.arcos_stats.value_changed.connect(
            self.update_event_counter
        )

    def update_event_counter(self):
        """Update the number of detected collective events."""
        df = self.data_storage_instance.arcos_stats.value

        if df.empty:
            self.widget.collev_number_display.display(0)
        else:
            collev_number = df["collid"].nunique()
            self.widget.collev_number_display.display(collev_number)


if __name__ == "__main__":
    import sys

    from arcos_gui.processing import DataStorage  # noqa: F811

    app = QtWidgets.QApplication(sys.argv)
    ds = DataStorage()
    controller = BottombarController(ds)
    controller.widget.show()
    sys.exit(app.exec_())



"""Widget to handle filtering of input data."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from arcos_gui.processing import filter_data, get_tracklengths
from arcos_gui.tools import (
    ARCOS_LAYERS,
    remove_layers_after_columnpicker,
    set_track_lenths,
)
from napari.utils.notifications import show_info
from qtpy import QtWidgets, uic
from qtpy.QtCore import Qt
from superqt import QRangeSlider

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage
    from napari.viewer import Viewer


class _filter_dataUI(QtWidgets.QWidget):
    UI_FILE = str(Path(__file__).parent.parent / "_ui" / "filter_data.ui")

    # The UI_FILE above contains these objects:
    frame_interval_label: QtWidgets.QLabel
    tracklength_label: QtWidgets.QLabel
    position_label: QtWidgets.QLabel
    position: QtWidgets.QComboBox
    additional_filter_combobox: QtWidgets.QComboBox
    additional_filter_combobox_label: QtWidgets.QLabel
    frame_interval: QtWidgets.QSpinBox
    min_tracklength: QtWidgets.QSlider
    min_tracklength_spinbox: QtWidgets.QDoubleSpinBox
    max_tracklength: QtWidgets.QSlider
    max_tracklength_spinbox: QtWidgets.QDoubleSpinBox
    filter_groupBox: QtWidgets.QGroupBox
    filter_input_data: QtWidgets.QPushButton
    horizontalLayout_tracklength: QtWidgets.QHBoxLayout

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui()

    def _init_ranged_sliderts(self):
        """Initialize ranged sliders from superqt."""
        self.tracklenght_slider = QRangeSlider(Qt.Horizontal)
        self.horizontalLayout_tracklength.addWidget(self.tracklenght_slider)

        # set starting values
        self.tracklenght_slider.setRange(0, 10)
        self.tracklenght_slider.setValue((0, 10))

    def _handle_slider_tracklength_value_change(self):
        """Method to handle trancklenght value changes."""
        slider_vals = self.tracklenght_slider.value()
        self.min_tracklength_spinbox.setValue(slider_vals[0])
        self.max_tracklength_spinbox.setValue(slider_vals[1])

    def _handle_min_tracklenght_box_value_change(self, value):
        """Method to handle min tracklenght spinbox."""
        slider_vals = self.tracklenght_slider.value()
        self.tracklenght_slider.setValue((value, slider_vals[1]))

    def _handle_max_tracklength_box_value_change(self, value):
        """Method to handle max tracklength spinbox."""
        slider_vals = self.tracklenght_slider.value()
        self.tracklenght_slider.setValue((slider_vals[0], value))

    def _connect_ranged_sliders_to_spinboxes(self):
        """Method to connect ranged sliders to spinboxes to sync values."""
        self.tracklenght_slider.valueChanged.connect(
            self._handle_slider_tracklength_value_change
        )
        self.min_tracklength_spinbox.valueChanged.connect(
            self._handle_min_tracklenght_box_value_change
        )
        self.max_tracklength_spinbox.valueChanged.connect(
            self._handle_max_tracklength_box_value_change
        )

    def _reset_filter_combobox(self):
        self.additional_filter_combobox.clear()
        self.position.clear()

    def set_defaults(self):
        """Method that sets the default visible widgets in the main window."""
        self.position.setVisible(False)
        self.position_label.setVisible(False)
        self.additional_filter_combobox.setVisible(False)
        self.additional_filter_combobox_label.setVisible(False)
        self._reset_filter_combobox()

    def set_position_visible(self):
        """Method that sets the visible widgets in the main window."""
        self.position.setVisible(True)
        self.position_label.setVisible(True)

    def set_additional_filter_visible(self):
        self.additional_filter_combobox.setVisible(True)
        self.additional_filter_combobox_label.setVisible(True)

    def setup_ui(self):
        """Setup UI. Loads it from ui file."""
        uic.loadUi(self.UI_FILE, self)  # load QtDesigner .ui file
        self.set_defaults()
        self._init_ranged_sliderts()
        self._connect_ranged_sliders_to_spinboxes()


class FilterController:
    """Widget to handle filtering of input data."""

    def __init__(self, viewer: Viewer, data_storage_instance: DataStorage, parent=None):
        self.viewer = viewer
        self.widget = _filter_dataUI()

        self.data_storage_instance = data_storage_instance
        self.data_storage_instance.original_data.value_changed.connect(
            self._original_data_changed
        )
        self.data_storage_instance.min_max_tracklenght.value_changed.connect(
            self._set_tracklengths
        )
        self.widget.max_tracklength_spinbox.valueChanged.connect(
            self._update_data_storage_tracklenght
        )
        self.widget.min_tracklength_spinbox.valueChanged.connect(
            self._update_data_storage_tracklenght
        )
        self.widget.filter_input_data.clicked.connect(self._filter_data)
        self._set_default_values()

    def _set_default_values(self):
        self.widget.set_defaults()
        self._set_tracklengths()

    def _filter_data(self):
        """Method to filter the data."""
        self._remove_old_layers()
        selected_position_value = self.widget.position.currentData()
        selected_additional_filter_value = (
            self.widget.additional_filter_combobox.currentData()
        )
        selected_frame_interval_value = self.widget.frame_interval.value()

        # get tracklengths
        min_tracklength = self.widget.min_tracklength_spinbox.value()
        max_tracklength = self.widget.max_tracklength_spinbox.value()

        # get input data
        input_data = self.data_storage_instance.original_data.value

        # get column names
        selected_fov_column = self.data_storage_instance.columns.value.position_id
        selected_object_id_column = self.data_storage_instance.columns.value.object_id
        selected_additional_filter_column = (
            self.data_storage_instance.columns.value.additional_filter_column
        )
        selected_frame_column = self.data_storage_instance.columns.value.frame_column
        measurement_name_column = (
            self.data_storage_instance.columns.value.measurement_column
        )
        coordinate_column = self.data_storage_instance.columns.value.posCol

        # filter data
        data_filtered, max_meas, min_meas = filter_data(
            df_in=input_data,
            field_of_view_id_name=selected_fov_column,
            frame_name=selected_frame_column,
            track_id_name=selected_object_id_column,
            measurement_name=measurement_name_column,
            additional_filter_column_name=selected_additional_filter_column,
            posCols=coordinate_column,
            fov_val=selected_position_value,
            additional_filter_value=selected_additional_filter_value,
            min_tracklength_value=min_tracklength,
            max_tracklength_value=max_tracklength,
            frame_interval=selected_frame_interval_value,
            st_out=show_info,
        )

        self._update_data_storage(data_filtered, min_meas, max_meas)

    def _original_data_changed(self):
        self._set_default_values()

        df_orig = self.data_storage_instance.original_data.value
        pos_col = self.data_storage_instance.columns.value.position_id
        add_filter_col = (
            self.data_storage_instance.columns.value.additional_filter_column
        )

        if pos_col:
            if len(df_orig[pos_col].unique()) > 1:
                self.widget.set_position_visible()
                for pos in df_orig[pos_col].unique():
                    self.widget.position.addItem(str(pos), pos)

        if add_filter_col:
            self.widget.set_additional_filter_visible()
            for add_filter in df_orig[add_filter_col].unique():
                self.widget.additional_filter_combobox.addItem(
                    str(add_filter), add_filter
                )

        self.widget.position.setCurrentIndex(0)
        self.widget.additional_filter_combobox.setCurrentIndex(0)
        self._filter_data()

    def _update_data_storage(self, df_filtered, min_meas, max_meas):
        """Method to update the data storage."""
        self.data_storage_instance.reset_relevant_attributes(True)
        self.data_storage_instance.filtered_data.value = df_filtered
        self.data_storage_instance.min_max_meas.value = [min_meas, max_meas]

    def _set_tracklengths(self):
        """Method to set the tracklengths."""
        if self.data_storage_instance.original_data.value.empty:
            min_t, max_t = self.data_storage_instance.min_max_tracklenght.value
        elif self.data_storage_instance.columns.value.object_id is None:
            min_t, max_t = 0, 1
        else:
            min_t, max_t = get_tracklengths(
                self.data_storage_instance.original_data.value,
                self.data_storage_instance.columns.value.position_id,
                self.data_storage_instance.columns.value.object_id,
                self.data_storage_instance.columns.value.additional_filter_column,
            )
        set_track_lenths(
            (min_t, max_t),
            self.widget.tracklenght_slider,
            self.widget.min_tracklength_spinbox,
            self.widget.max_tracklength_spinbox,
        )

    def _remove_old_layers(self):
        remove_layers_after_columnpicker(self.viewer, ARCOS_LAYERS.values())

    def _update_data_storage_tracklenght(self):
        """Method to update the data storage."""
        self.data_storage_instance.toggle_callback_block(True)
        self.data_storage_instance.min_max_tracklenght.value = [
            self.widget.min_tracklength_spinbox.value(),
            self.widget.max_tracklength_spinbox.value(),
        ]
        self.data_storage_instance.toggle_callback_block(False)



"""Widget to set arcos parameters and run arcos algorithm."""

from __future__ import annotations

import traceback
from dataclasses import fields
from pathlib import Path
from typing import TYPE_CHECKING

from arcos_gui.processing import arcos_worker
from arcos_gui.tools import OutputOrderValidator
from napari.utils import progress
from napari.utils.notifications import show_info
from qtpy import QtWidgets, uic
from qtpy.QtCore import QSize, QTimer, Signal
from qtpy.QtGui import QIcon, QMovie, QValidator

if TYPE_CHECKING:
    from arcos_gui.processing import DataStorage

# icons
ICONS = Path(__file__).parent.parent / "_icons"


class _arcosWidget(QtWidgets.QWidget):
    closing = Signal()
    UI_FILE = str(Path(__file__).parent.parent / "_ui" / "ARCOS_widget.ui")

    # The UI_FILE above contains these objects:

    clip_low_label: QtWidgets.QLabel
    clip_high_label: QtWidgets.QLabel
    intervale_type_label: QtWidgets.QLabel
    bias_method_label: QtWidgets.QLabel
    smooth_k_label: QtWidgets.QLabel
    bias_k_label: QtWidgets.QLabel
    polyDeg_label: QtWidgets.QLabel
    bin_peak_threshold_label: QtWidgets.QLabel
    bin_threshold_label: QtWidgets.QLabel
    neighbourhood_label: QtWidgets.QLabel
    min_clustersize_label: QtWidgets.QLabel
    nprev_spinbox: QtWidgets.QSpinBox
    nprev_label: QtWidgets.QLabel
    min_dur_label: QtWidgets.QLabel
    tot_size_label: QtWidgets.QLabel

    interpolate_meas: QtWidgets.QCheckBox
    clip_meas: QtWidgets.QCheckBox
    clip_low: QtWidgets.QDoubleSpinBox
    clip_high: QtWidgets.QDoubleSpinBox

    bin_advanced_options: QtWidgets.QCheckBox
    bias_method: QtWidgets.QComboBox
    smooth_k: QtWidgets.QSpinBox
    bias_k: QtWidgets.QSpinBox
    polyDeg: QtWidgets.QSpinBox
    bin_peak_threshold: QtWidgets.QDoubleSpinBox
    bin_threshold: QtWidgets.QDoubleSpinBox

    detect_advance_options: QtWidgets.QCheckBox
    eps_estimation_label: QtWidgets.QLabel
    eps_estimation_combobox: QtWidgets.QComboBox
    neighbourhood_size: QtWidgets.QDoubleSpinBox
    Cluster_linking_dist_checkbox: QtWidgets.QCheckBox
    epsPrev_spinbox: QtWidgets.QDoubleSpinBox
    min_clustersize: QtWidgets.QSpinBox
    min_dur: QtWidgets.QSpinBox
    total_event_size: QtWidgets.QSpinBox
    update_arcos: QtWidgets.QPushButton
    run_binarization_only: QtWidgets.QPushButton
    arcos_group: QtWidgets.QGroupBox
    clip_frame: QtWidgets.QFrame
    add_all_cells_checkbox: QtWidgets.QCheckBox
    add_bin_cells_checkbox: QtWidgets.QCheckBox
    add_convex_hull_checkbox: QtWidgets.QCheckBox
    output_order: QtWidgets.QLineEdit
    output_order_label: QtWidgets.QLabel
    loading_label: QtWidgets.QLabel

    cancel_button: QtWidgets.QPushButton

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setup_ui()

    def setup_ui(self):
        """Setup UI. Loads it from ui file."""
        uic.loadUi(self.UI_FILE, self)  # load QtDesigner .ui file
        self.loading_icon = QMovie(str(ICONS / "Dual Ring-1s-200px.gif"))
        self.loading_icon.setScaledSize(QSize(40, 40))
        self.loading_label.setMovie(self.loading_icon)
        self.loading_icon.start()
        # self.loading_icon.stop()
        self.loading_label.hide()
        self.cancel_button.setStyleSheet("background-color : #7C0A02; color : white")
        self.cancel_button.hide()
        self._set_advanced_state_dict()
        self._setup_timer()
        self._connect_ui_callbacks()
        self._init_callbacks_visible_arcosparameters()
        self._set_default_visible()
        self.updateValidator(None)

    def _set_advanced_state_dict(self):
        self.set_default_bin_state_dict()
        self._set_detect_advanced_state_dict()

    def set_default_bin_state_dict(self):
        self.bias_met_advanced_state = {
            "bias_method": self.bias_method.currentText(),
            "smoothK": self.smooth_k.value(),
        }

    def _set_detect_advanced_state_dict(self):
        self.detect_advanced_state = {
            "eps": self.neighbourhood_size.value(),
            "auto_eps": self.eps_estimation_combobox.currentText(),
            "cluter_linking_dist_check": self.Cluster_linking_dist_checkbox.isChecked(),
            "epsPrev": self.epsPrev_spinbox.value(),
            "nPrev": self.nprev_spinbox.value(),
        }

    def _connect_ui_callbacks(self):
        self.output_order.textChanged.connect(self._onTextChanged)
        self.bin_advanced_options.stateChanged.connect(
            self._bin_advanced_options_toggle
        )
        self.detect_advance_options.stateChanged.connect(
            self._detect_advanced_options_toggle
        )
        self.eps_estimation_combobox.currentIndexChanged.connect(
            self._eps_estimation_toggle
        )
        self.Cluster_linking_dist_checkbox.stateChanged.connect(self._epsPrev_toggle)

    def _onTextChanged(self, text):
        validator = self.output_order.validator()
        if validator:
            state, _, _ = validator.validate(text.lower(), len(text))
            if state == QValidator.Acceptable:
                self.update_arcos.setEnabled(True)
                self.run_binarization_only.setEnabled(True)
                self.output_order.setStyleSheet("")
            elif state == QValidator.Intermediate or state == QValidator.Invalid:
                color = "#f6989d"  # light red
                self.update_arcos.setEnabled(False)
                self.run_binarization_only.setEnabled(False)
                self.output_order.setStyleSheet(
                    "QLineEdit { background-color: %s; color: black }" % color
                )

    def updateValidator(self, vColsCore: list[str] | None = None):
        if vColsCore is None:
            vColsCore = ["t", "x", "y"]
        validator = OutputOrderValidator(vColsCore, self.output_order)
        self.output_order.setValidator(validator)
        if len(vColsCore) == 3:
            self.output_order.setText("tyx")
        elif len(vColsCore) == 4:
            self.output_order.setText("tzyx")

    def _set_default_visible(self):
        """Method that sets the default visible widgets in the main window."""
        self.clip_meas.setChecked(False)
        self._toggle_bias_method_parameter_visibility()
        self._bin_advanced_options_toggle()
        self._detect_advanced_options_toggle()
        self._epsPrev_toggle()
        self._eps_estimation_toggle()

    def _bin_advanced_options_toggle(self):
        checked = self.bin_advanced_options.isChecked()
        if checked:
            self.bias_method.setCurrentText(self.bias_met_advanced_state["bias_method"])
            self.smooth_k.setValue(self.bias_met_advanced_state["smoothK"])
        else:
            self.bias_met_advanced_state["bias_method"] = self.bias_method.currentText()
            self.bias_met_advanced_state["smoothK"] = self.smooth_k.value()
            self.smooth_k.setValue(3)
            self.bias_method.setCurrentText("none")
            self.bias_method.currentIndexChanged.emit(0)

        self._toggle_bias_method_parameter_visibility()

    def _detect_advanced_options_toggle(self):
        checked = self.detect_advance_options.isChecked()
        self.eps_estimation_label.setVisible(checked)
        self.eps_estimation_combobox.setVisible(checked)
        self.neighbourhood_label.setVisible(checked)
        self.neighbourhood_size.setVisible(checked)
        self.Cluster_linking_dist_checkbox.setVisible(checked)
        self.epsPrev_spinbox.setVisible(checked)
        self.nprev_label.setVisible(checked)
        self.nprev_spinbox.setVisible(checked)
        if checked:
            self.eps_estimation_combobox.setCurrentText(
                self.detect_advanced_state["auto_eps"]
            )
            self.neighbourhood_size.setValue(self.detect_advanced_state["eps"])
            self.Cluster_linking_dist_checkbox.setChecked(
                self.detect_advanced_state["cluter_linking_dist_check"]
            )
            self.epsPrev_spinbox.setValue(self.detect_advanced_state["epsPrev"])
            self.nprev_spinbox.setValue(self.detect_advanced_state["nPrev"])
        else:
            self.detect_advanced_state = {
                "eps": self.neighbourhood_size.value(),
                "auto_eps": self.eps_estimation_combobox.currentText(),
                "cluter_linking_dist_check": self.Cluster_linking_dist_checkbox.isChecked(),
                "epsPrev": self.epsPrev_spinbox.value(),
                "nPrev": self.nprev_spinbox.value(),
            }
            self.eps_estimation_combobox.setCurrentText("mean")
            self.Cluster_linking_dist_checkbox.setChecked(False)
            self.nprev_spinbox.setValue(1)

    def _eps_estimation_toggle(self):
        eps_method = self.eps_estimation_combobox.currentText()
        if eps_method == "manual":
            self.neighbourhood_size.setEnabled(True)
            self.neighbourhood_size.setButtonSymbols(
                QtWidgets.QAbstractSpinBox.UpDownArrows
            )
        else:
            self.neighbourhood_size.setEnabled(False)
            self.neighbourhood_size.setButtonSymbols(
                QtWidgets.QAbstractSpinBox.NoButtons
            )

    def _epsPrev_toggle(self):
        checked = self.Cluster_linking_dist_checkbox.isChecked()
        self.epsPrev_spinbox.setEnabled(checked)
        if checked:
            self.neighbourhood_size.valueChanged.disconnect(
                self._update_epsPrev_from_eps
            )
            self.epsPrev_spinbox.setButtonSymbols(
                QtWidgets.QAbstractSpinBox.UpDownArrows
            )
        else:
            self._update_epsPrev_from_eps()
            self.neighbourhood_size.valueChanged.connect(self._update_epsPrev_from_eps)
            self.epsPrev_spinbox.setButtonSymbols(QtWidgets.QAbstractSpinBox.NoButtons)

    def _update_epsPrev_from_eps(self):
        self.epsPrev_spinbox.setValue(self.neighbourhood_size.value())

    def _toggle_bias_method_parameter_visibility(self):
        """
        based on the seleciton of bias method:
        shows or hides the appropriate options in the main window.
        """
        advanced_checked = self.bin_advanced_options.isChecked()
        self.smooth_k.setVisible(advanced_checked)
        self.smooth_k_label.setVisible(advanced_checked)
        self.bias_method.setVisible(advanced_checked)
        self.bias_method_label.setVisible(advanced_checked)

        if self.bias_method.currentText() == "runmed":
            self.polyDeg.setVisible(False)
            self.polyDeg_label.setVisible(False)
            self.bias_k.setVisible(True)
            self.bias_k_label.setVisible(True)
            self.bin_peak_threshold.setVisible(True)
            self.bin_peak_threshold_label.setVisible(True)
            self.bin_threshold.setVisible(True)
            self.bin_threshold_label.setVisible(True)

        if self.bias_method.currentText() == "lm":
            self.polyDeg.setVisible(True)
            self.polyDeg_label.setVisible(True)
            self.bias_k.setVisible(False)
            self.bias_k_label.setVisible(False)
            self.bin_peak_threshold.setVisible(True)
            self.bin_peak_threshold_label.setVisible(True)
            self.bin_threshold.setVisible(True)
            self.bin_threshold_label.setVisible(True)

        if self.bias_method.currentText() == "none":
            self.polyDeg.setVisible(False)
            self.polyDeg_label.setVisible(False)
            self.bias_k.setVisible(False)
            self.bias_k_label.setVisible(False)
            self.bin_peak_threshold.setVisible(False)
            self.bin_peak_threshold_label.setVisible(False)
            self.bin_threshold.setVisible(True)
            self.bin_threshold_label.setVisible(True)

    def toggle_bias_method_enable(self, enabled: bool):
        """Toggle the enable of the bias method groupbox."""
        self.bias_method.setEnabled(enabled)
        self.bias_method_label.setEnabled(enabled)
        self.smooth_k.setEnabled(enabled)
        self.smooth_k_label.setEnabled(enabled)
        self.polyDeg.setEnabled(enabled)
        self.polyDeg_label.setEnabled(enabled)
        self.bias_k.setEnabled(enabled)
        self.bias_k_label.setEnabled(enabled)
        self.bin_peak_threshold.setEnabled(enabled)
        self.bin_peak_threshold_label.setEnabled(enabled)

    def toggle_tracklenght_filter_enable(self, enabled: bool):
        self.total_event_size.setEnabled(enabled)
        self.tot_size_label.setEnabled(enabled)

    def _toggle_clip_visible(self):
        """Toggle the visibility of the clipping options."""
        if self.clip_meas.isChecked():
            self.clip_high.setVisible(True)
            self.clip_low.setVisible(True)
            self.clip_high_label.setVisible(True)
            self.clip_low_label.setVisible(True)
        else:
            self.clip_high.setVisible(False)
            self.clip_low.setVisible(False)
            self.clip_high_label.setVisible(False)
            self.clip_low_label.setVisible(False)

    def _init_callbacks_visible_arcosparameters(self):
        """Initialize the callbacks for visible parameters in bias method groupbox."""
        self.bias_method.currentIndexChanged.connect(
            self._toggle_bias_method_parameter_visibility
        )
        self.clip_meas.stateChanged.connect(self._toggle_clip_visible)

    def _set_loading_icon(self, frame=None):
        self.update_arcos.setIcon(QIcon(self.loading_icon.currentPixmap()))

    def _hide_loading_icon(self):
        self.loading_label.hide()
        self.loading_icon.stop()

    def _setup_timer(self):
        self.timer_loading_icon_show = QTimer(parent=self)
        self.timer_loading_icon_show.setSingleShot(True)

    def start_loading(self):
        """Start loading icon animation."""
        self.update_arcos.setEnabled(False)
        self.run_binarization_only.setEnabled(False)
        self.cancel_button.show()
        self.cancel_button.setEnabled(True)
        self.timer_loading_icon_show.timeout.connect(self._show_loading_icon)
        self.timer_loading_icon_show.start(100)

    def _show_loading_icon(self):
        self.loading_label.show()
        self.loading_icon.start()

    def stop_loading(self):
        """Stop loading icon animation."""
        if self.timer_loading_icon_show.isActive():
            self.timer_loading_icon_show.stop()
        self._hide_loading_icon()
        self.cancel_button.hide()
        self.cancel_button.setEnabled(False)
        self.update_arcos.setEnabled(True)
        self.run_binarization_only.setEnabled(True)

    def closeEvent(self, event):
        self.closing.emit()
        event.accept()


class ArcosController:
    """Widget to set arcos parameters and run arcos algorithm."""

    def __init__(self, data_storage_instance: DataStorage, parent=None):
        self.widget = _arcosWidget(parent)

        self._what_to_run: set = set()
        self._data_storage_instance = data_storage_instance
        self.abort_timer = QTimer(parent)
        self.abort_timer.timeout.connect(self.abort_timer_timeout)

        self._init_callbacks_for_whattorun()
        self._update_what_to_run_all()
        self._update_arcos_parameters_object()
        self._connect_callbacks()

    def _connect_callbacks(self):
        self.widget.closing.connect(self.closeEvent)
        self.widget.update_arcos.clicked.connect(self._run_arcos)
        self.widget.run_binarization_only.clicked.connect(self._run_binarization_only)

        self.widget.cancel_button.clicked.connect(self.abort_worker)
        self._data_storage_instance.original_data.value_changed.connect(
            self._update_bias_method_availability
        )
        self._data_storage_instance.original_data.value_changed.connect(
            self._update_filter_availability
        )
        self._data_storage_instance.original_data.value_changed.connect(
            lambda: self.widget.updateValidator(
                self._data_storage_instance.columns.value.vcolscore
            )
        )

        self._connect_arcos_parameter_callbacks()

    def _update_bias_method_availability(self):
        """Change availability of bias methods depending on input data."""
        if self._data_storage_instance.columns.value.object_id is None:
            self.widget.toggle_bias_method_enable(False)
            self.widget.bias_method.setCurrentIndex(0)
        else:
            self.widget.toggle_bias_method_enable(True)

    def _update_filter_availability(self):
        """Change availability of filter methods depending on input data."""
        if self._data_storage_instance.columns.value.object_id is None:
            self.widget.toggle_tracklenght_filter_enable(False)
        else:
            self.widget.toggle_tracklenght_filter_enable(True)

    def _update_datastorage_with_bin_data(self, bin_data):
        self._data_storage_instance.columns.value.measurement_bin = bin_data[0]
        self._data_storage_instance.columns.value.measurement_resc = bin_data[1]
        self._data_storage_instance.arcos_binarization.value = bin_data[2]

    def _update_datastorage_with_arcos(self, arcos_data):
        self._data_storage_instance.arcos_stats.value = arcos_data[1]
        self._data_storage_instance.arcos_output.value = arcos_data[0]

    def _setup_debounce_timer(self):
        self.debounce_time = (
            10  # Adjust this value to set the debounce time in milliseconds
        )
        self.missed_updates = 0
        self.update_timer = QTimer()
        self.update_timer.setSingleShot(True)
        self.update_timer.timeout.connect(self._perform_update)

    def _perform_update(self):
        # Get the current value of the progress bar
        current_value = self.pbar.n

        # Update the progress bar with the missed updates
        self.pbar.update(self.missed_updates)

        # Correct the number of missed updates based on the actual change in the progress bar value
        self.missed_updates = (current_value + self.missed_updates) - self.pbar.n

        # Reset the missed updates counter if it's zero
        if self.pbar.n >= self.pbar.total:
            self.missed_updates = 0

    def createWorkerThread(self, what_to_run: set):
        """Create a worker thread to run arcos algorithm."""
        # try to get the old ARCOS object from the worker,
        # this allows to continue the calculation if the worker is reset
        try:
            self.worker: arcos_worker  # for mypy
            arcos_object = self.worker.arcos_object
            arcos_raw_output = self.worker.arcos_raw_output
        except AttributeError:
            arcos_object = None
            arcos_raw_output = None

        self._setup_debounce_timer()

        self.pbar = progress(total=0, desc="ARCOS calculation")

        self.worker = arcos_worker(
            what_to_run,
            show_info,
            self._data_storage_instance.arcos_parameters.value,
            columns=self._data_storage_instance.columns.value,
            filtered_data=self._data_storage_instance.filtered_data.value,
            arcos_object=arcos_object,
            arcos_raw_output=arcos_raw_output,
        )
        self.worker.arcos_progress_update.connect(self._update_progressbar)

        self.worker.binarization_finished.connect(
            self._update_datastorage_with_bin_data
        )
        self.worker.new_eps.connect(self._update_eps)
        self.worker.new_arcos_output.connect(self._update_datastorage_with_arcos)
        self.worker.finished.connect(self.widget.stop_loading)
        self.worker.started.connect(self.widget.start_loading)
        self.worker.finished.connect(self.abort_timer_stop)
        self.worker.errored.connect(self.abort_timer_stop)
        self.worker.errored.connect(self._show_errored_message)

        self.worker.start()
        return self.worker

    def _update_progressbar(self, update_type, value):
        if update_type == "total":
            self.pbar.reset(value)
        elif update_type == "reset":
            self.pbar.reset(value)
        elif update_type == "update":
            self.missed_updates += value
            if not self.update_timer.isActive():
                self.update_timer.start(self.debounce_time)
        else:
            raise ValueError(f"Unknown update_type: {update_type}")

    def _run_arcos(self):
        self.createWorkerThread(self._what_to_run)

    def _run_binarization_only(self):
        worker = self.createWorkerThread({"binarization"})
        worker.finished.connect(self._if_data_update_what_to_run)

    def _if_data_update_what_to_run(self):
        if not self._data_storage_instance.arcos_binarization.value.empty:
            self._update_what_to_run_tracking()

    def _show_errored_message(self, err):
        show_info(f"ARCOS calculation aborted due to: {err}")
        traceback.print_exception(None, err, err.__traceback__)

    def _update_eps(self, eps):
        self._data_storage_instance.arcos_parameters.value.neighbourhood_size.value = (
            eps
        )

    def update_worker_with_data(self):
        self.worker.what_to_run = self._what_to_run
        self.worker.filtered_data = self._data_storage_instance.filtered_data.value
        self.worker.columns = self._data_storage_instance.columns.value
        self.worker.arcos_parameters.set_all_parameters(
            self._data_storage_instance.arcos_parameters.value
        )

    def abort_worker(self):
        self.worker.finished.connect(self.abort_timer_stop)
        self.worker.quit()
        self.widget.cancel_button.setEnabled(False)
        self.widget.cancel_button.setText("Aborting")
        self.abort_timer_start()

    def abort_timer_start(self):
        self.abort_timer.start(1000)

    def abort_timer_timeout(self):
        self.update_abort_button()

    def abort_timer_stop(self):
        self.abort_timer.stop()
        self.pbar.close()
        self.widget.cancel_button.setText("Abort")

    def update_abort_button(self):
        self.dots = self.widget.cancel_button.text().count(".")
        if self.dots < 3:
            self.widget.cancel_button.setText(self.widget.cancel_button.text() + ".")
        else:
            self.widget.cancel_button.setText("Aborting")

    def _stop_worker(self):
        try:
            self.worker.quit()
        except (AttributeError, RuntimeError):
            pass

    def closeEvent(self):
        self._stop_worker()

    def _init_callbacks_for_whattorun(self):
        """Connect callbacks for updating 'what to run'.

        This is used to run only the appropriate parts of the
        run_arcos method depending on the parameters changed.
        """
        # for every changable field connect the appropriate what_to_run function
        for i in [self.widget.bias_method]:
            i.currentIndexChanged.connect(self._update_what_to_run_all)
        for i in [
            self.widget.clip_low,
            self.widget.clip_high,
            self.widget.smooth_k,
            self.widget.bias_k,
            self.widget.polyDeg,
            self.widget.bin_threshold,
            self.widget.bin_peak_threshold,
        ]:
            i.valueChanged.connect(self._update_what_to_run_all)
        for i in [self.widget.interpolate_meas, self.widget.clip_meas]:
            i.stateChanged.connect(self._update_what_to_run_all)
        for i in [
            self.widget.neighbourhood_size,
            self.widget.min_clustersize,
            self.widget.nprev_spinbox,
        ]:
            i.valueChanged.connect(self._update_what_to_run_tracking)
        for i in [self.widget.epsPrev_spinbox]:
            i.valueChanged.connect(self._update_what_to_run_tracking)
        for i in [self.widget.eps_estimation_combobox]:
            i.currentIndexChanged.connect(self._update_what_to_run_tracking)
        for i in [self.widget.min_dur, self.widget.total_event_size]:
            i.valueChanged.connect(self._update_what_to_run_filtering)
        for i in [
            self.widget.add_convex_hull_checkbox,
            self.widget.add_all_cells_checkbox,
            self.widget.add_bin_cells_checkbox,
        ]:
            i.stateChanged.connect(self._update_what_to_run_filtering)

        self.widget.output_order.textChanged.connect(self._update_what_to_run_filtering)

    def _update_what_to_run_all(self):
        """Adds 'all' to the what_to_run attirbute,
        that is used in the main function to check if what to run
        when certain field have updated values."""
        self._clear_what_to_run()
        self._what_to_run.add("binarization")
        self._what_to_run.add("tracking")
        self._what_to_run.add("filtering")

    def _update_what_to_run_tracking(self):
        """Adds 'from_tracking' to the what_to_run attirbute,
        that is used in the main function to check if what to run
        when certain field have updated values."""
        self._what_to_run.add("tracking")
        self._what_to_run.add("filtering")

    def _update_what_to_run_filtering(self):
        """Adds 'from_filtering' to the what_to_run attirbute,
        that is used in the main function to check if what to run
        when certain field have updated values."""
        self._what_to_run.add("filtering")

    def _clear_what_to_run(self):
        """Clears the what_to_run attribute."""
        self._what_to_run.clear()

    def _update_arcos_parameters_object(self):
        """Update the parameters in the data storage instance"""
        try:
            self._data_storage_instance.toggle_callback_block(True)
            if not self.widget.Cluster_linking_dist_checkbox.isChecked():
                epsPrev = None
            else:
                epsPrev = self.widget.epsPrev_spinbox.value()
            self._data_storage_instance.arcos_parameters.value.interpolate_meas.value = (
                self.widget.interpolate_meas.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.clip_meas.value = (
                self.widget.clip_meas.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.clip_low.value = (
                self.widget.clip_low.value()
            )
            self._data_storage_instance.arcos_parameters.value.clip_high.value = (
                self.widget.clip_high.value()
            )
            self._data_storage_instance.arcos_parameters.value.bin_advanded_settings.value = (
                self.widget.bin_advanced_options.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.smooth_k.value = (
                self.widget.smooth_k.value()
            )
            self._data_storage_instance.arcos_parameters.value.bias_k.value = (
                self.widget.bias_k.value()
            )
            self._data_storage_instance.arcos_parameters.value.bias_method.value = (
                self.widget.bias_method.currentText()
            )
            self._data_storage_instance.arcos_parameters.value.polyDeg.value = (
                self.widget.polyDeg.value()
            )
            self._data_storage_instance.arcos_parameters.value.bin_threshold.value = (
                self.widget.bin_threshold.value()
            )
            self._data_storage_instance.arcos_parameters.value.bin_peak_threshold.value = (
                self.widget.bin_peak_threshold.value()
            )
            self._data_storage_instance.arcos_parameters.value.detect_advanced_options.value = (
                self.widget.detect_advance_options.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.neighbourhood_size.value = (
                self.widget.neighbourhood_size.value()
            )
            self._data_storage_instance.arcos_parameters.value.min_clustersize.value = (
                self.widget.min_clustersize.value()
            )
            self._data_storage_instance.arcos_parameters.value.nprev.value = (
                self.widget.nprev_spinbox.value()
            )
            self._data_storage_instance.arcos_parameters.value.min_dur.value = (
                self.widget.min_dur.value()
            )
            self._data_storage_instance.arcos_parameters.value.total_event_size.value = (
                self.widget.total_event_size.value()
            )
            self._data_storage_instance.arcos_parameters.value.add_convex_hull.value = (
                self.widget.add_convex_hull_checkbox.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.add_all_cells.value = (
                self.widget.add_all_cells_checkbox.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.add_bin_cells.value = (
                self.widget.add_bin_cells_checkbox.isChecked()
            )
            self._data_storage_instance.arcos_parameters.value.epsPrev.value = epsPrev
            self._data_storage_instance.arcos_parameters.value.eps_method.value = (
                self.widget.eps_estimation_combobox.currentText()
            )
            self._data_storage_instance.output_order.value = (
                self.widget.output_order.text()
            )
        finally:
            self._data_storage_instance.toggle_callback_block(False)

    def _update_arcos_parameters_ui(self):
        try:
            self._data_storage_instance.toggle_callback_block(True)
            self.block_qt_signals(True)

            self.widget.interpolate_meas.setChecked(
                self._data_storage_instance.arcos_parameters.value.interpolate_meas.value
            )
            self.widget.clip_meas.setChecked(
                self._data_storage_instance.arcos_parameters.value.clip_meas.value
            )
            self.widget.clip_low.setValue(
                self._data_storage_instance.arcos_parameters.value.clip_low.value
            )
            self.widget.clip_high.setValue(
                self._data_storage_instance.arcos_parameters.value.clip_high.value
            )
            self.widget.smooth_k.setValue(
                self._data_storage_instance.arcos_parameters.value.smooth_k.value
            )
            self.widget.bias_k.setValue(
                self._data_storage_instance.arcos_parameters.value.bias_k.value
            )
            self.widget.bias_method.setCurrentText(
                self._data_storage_instance.arcos_parameters.value.bias_method.value
            )
            self.widget.polyDeg.setValue(
                self._data_storage_instance.arcos_parameters.value.polyDeg.value
            )
            self.widget.bin_threshold.setValue(
                self._data_storage_instance.arcos_parameters.value.bin_threshold.value
            )
            self.widget.bin_peak_threshold.setValue(
                self._data_storage_instance.arcos_parameters.value.bin_peak_threshold.value
            )
            self.widget.neighbourhood_size.setValue(
                self._data_storage_instance.arcos_parameters.value.neighbourhood_size.value
            )
            self.widget.min_clustersize.setValue(
                self._data_storage_instance.arcos_parameters.value.min_clustersize.value
            )
            self.widget.nprev_spinbox.setValue(
                self._data_storage_instance.arcos_parameters.value.nprev.value
            )
            self.widget.min_dur.setValue(
                self._data_storage_instance.arcos_parameters.value.min_dur.value
            )
            self.widget.total_event_size.setValue(
                self._data_storage_instance.arcos_parameters.value.total_event_size.value
            )
            self.widget.add_convex_hull_checkbox.setChecked(
                self._data_storage_instance.arcos_parameters.value.add_convex_hull.value
            )
            self.widget.add_all_cells_checkbox.setChecked(
                self._data_storage_instance.arcos_parameters.value.add_all_cells.value
            )
            self.widget.add_bin_cells_checkbox.setChecked(
                self._data_storage_instance.arcos_parameters.value.add_bin_cells.value
            )
            self.widget.eps_estimation_combobox.setCurrentText(
                self._data_storage_instance.arcos_parameters.value.eps_method.value
            )
            if (
                self._data_storage_instance.arcos_parameters.value.epsPrev.value
                is not None
            ):
                self.widget.Cluster_linking_dist_checkbox.setChecked(True)
                self.widget.epsPrev_spinbox.setValue(
                    self._data_storage_instance.arcos_parameters.value.epsPrev.value
                )
            else:
                self.widget.Cluster_linking_dist_checkbox.setChecked(False)
                self.widget.epsPrev_spinbox.setValue(
                    self._data_storage_instance.arcos_parameters.value.neighbourhood_size.value
                )

            self.widget.output_order.setText(
                self._data_storage_instance.output_order.value
            )
            self.widget.bin_advanced_options.setChecked(
                self._data_storage_instance.arcos_parameters.value.bin_advanded_settings.value
            )
            self.widget.detect_advance_options.setChecked(
                self._data_storage_instance.arcos_parameters.value.detect_advanced_options.value
            )
            self.widget._set_advanced_state_dict()
        finally:
            self._data_storage_instance.toggle_callback_block(False)
            self.block_qt_signals(False)
            self.widget._bin_advanced_options_toggle()
            self.widget._detect_advanced_options_toggle()

    def _connect_arcos_parameter_callbacks(self):
        for param in fields(self._data_storage_instance.arcos_parameters.value):
            getattr(
                self._data_storage_instance.arcos_parameters.value, param.name
            ).value_changed.connect(self._update_arcos_parameters_ui)

        self._data_storage_instance.output_order.value_changed.connect(
            self._update_arcos_parameters_ui
        )

        # whenever the ui is updated, update the data storage instance
        for i in [self.widget.bias_method]:
            i.currentIndexChanged.connect(self._update_arcos_parameters_object)
        for i in [
            self.widget.clip_low,
            self.widget.clip_high,
            self.widget.smooth_k,
            self.widget.bias_k,
            self.widget.polyDeg,
            self.widget.bin_threshold,
            self.widget.bin_peak_threshold,
        ]:
            i.valueChanged.connect(self._update_arcos_parameters_object)
        for i in [
            self.widget.interpolate_meas,
            self.widget.clip_meas,
            self.widget.bin_advanced_options,
            self.widget.detect_advance_options,
        ]:
            i.stateChanged.connect(self._update_arcos_parameters_object)
        for i in [
            self.widget.neighbourhood_size,
            self.widget.min_clustersize,
            self.widget.nprev_spinbox,
        ]:
            i.valueChanged.connect(self._update_arcos_parameters_object)
        for i in [self.widget.epsPrev_spinbox]:
            i.valueChanged.connect(self._update_arcos_parameters_object)
        for i in [self.widget.eps_estimation_combobox]:
            i.currentIndexChanged.connect(self._update_arcos_parameters_object)
        for i in [self.widget.min_dur, self.widget.total_event_size]:
            i.valueChanged.connect(self._update_arcos_parameters_object)
        for i in [
            self.widget.add_convex_hull_checkbox,
            self.widget.add_all_cells_checkbox,
            self.widget.add_bin_cells_checkbox,
        ]:
            i.stateChanged.connect(self._update_arcos_parameters_object)

        self.widget.output_order.textChanged.connect(
            self._update_arcos_parameters_object
        )

    def block_qt_signals(self, block=True):
        """Block or unblock signals for a widget and all its child widgets."""
        self.widget.blockSignals(block)
        for child_widget in self.widget.findChildren(QtWidgets.QWidget):
            child_widget.blockSignals(block)

    def __del__(self):
        try:
            self._stop_worker()
        except (AttributeError, RuntimeError):
            pass


if __name__ == "__main__":
    import sys

    from arcos_gui.processing import DataStorage  # noqa: F811

    app = QtWidgets.QApplication(sys.argv)
    controller = ArcosController(DataStorage())
    controller.widget.show()
    sys.exit(app.exec_())



"""Exporting widget for arcos-gui."""
from __future__ import annotations

import os
import traceback
import warnings
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING

from arcos_gui.processing import BatchProcessor
from arcos_gui.tools import (
    ALLOWED_SETTINGS,
    AVAILABLE_OPTIONS_FOR_BATCH,
    BatchFileDialog,
    MovieExporter,
    ParameterFileDialog,
)
from napari.utils import progress
from napari.utils.notifications import show_info
from qtpy import QtWidgets, uic
from qtpy.QtCore import QTimer, Signal
from qtpy.QtGui import QIcon

if TYPE_CHECKING:
    import napari.viewer
    from arcos_gui.processing import DataStorage

# icons
ICONS = Path(__file__).parent.parent / "_icons"


class _exportwidget(QtWidgets.QWidget):
    UI_FILE = str(Path(__file__).parent.parent / "_ui" / "export_widget.ui")

    # The UI_FILE above contains these objects:

    browse_file_data: QtWidgets.QPushButton
    file_LineEdit_data: QtWidgets.QLineEdit
    base_name_LineEdit_data: QtWidgets.QLineEdit
    data_export_button: QtWidgets.QPushButton
    param_export_button: QtWidgets.QPushButton
    param_import_button: QtWidgets.QPushButton
    stats_export_button: QtWidgets.QPushButton
    batch_processing_button: QtWidgets.QPushButton
    abort_batch_button: QtWidgets.QPushButton

    browse_file_img: QtWidgets.QPushButton
    file_LineEdit_img: QtWidgets.QLineEdit
    base_name_LineEdit_img: QtWidgets.QLineEdit
    img_seq_export_button: QtWidgets.QPushButton
    automatic_size_img: QtWidgets.QCheckBox
    spinBox_height_img: QtWidgets.QSpinBox
    spinBox_width_img: QtWidgets.QSpinBox

    closing = Signal()

    def __init__(self, data_storage: DataStorage, parent=None):
        super().__init__(parent)
        self._data_storage_instance = data_storage
        self._batch_process_path: str | None = None
        self._parameters_path: str | None = None
        self.setup_ui()

    def _browse_file_data(self):
        base_path = self._data_storage_instance.file_name.value
        base_path = str(Path(base_path).parent)
        path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select Directory", base_path
        )
        self.file_LineEdit_data.setText(path)

    def _browse_file_img(self):
        base_path = self._data_storage_instance.file_name.value
        base_path = str(Path(base_path).parent)
        path = QtWidgets.QFileDialog.getExistingDirectory(
            self, "Select Directory", base_path
        )
        self.file_LineEdit_img.setText(path)

    def _browse_parmeters_export(self):
        if self._parameters_path is None:
            base_path = self._data_storage_instance.file_name.value
        else:
            base_path = self._parameters_path
        base_path = str(Path(base_path).parent)
        path = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save as Yaml", base_path, filter="*.yaml"
        )[0]
        self._parameters_path = path
        if path:
            if not path.endswith(".yaml"):
                path = os.path.join(path, ".yaml")
            return path
        else:
            return None

    def _browse_parmeters_import(self):
        if self._parameters_path is None:
            base_path = self._data_storage_instance.file_name.value
        else:
            base_path = self._parameters_path
        base_path = str(Path(base_path).parent)

        dialog = ParameterFileDialog(
            selection_values=ALLOWED_SETTINGS,
            directory=base_path,
            parent=self,
            caption="Select Parameters to Import",
        )

        if dialog.exec_():
            # Get the selected directory
            path = dialog.selectedFiles()[0]
            self._parameters_path = path

            # Get the values corresponding to the checkboxes that are checked
            options_selected = dialog.get_selected_options()
            dialog.close()
            dialog.deleteLater()
            return path, options_selected
        else:
            dialog.close()
            dialog.deleteLater()
            return None, None

    def _browse_batch_output(self):
        if self._batch_process_path is None:
            base_path = self._data_storage_instance.file_name.value
        else:
            base_path = self._batch_process_path
        base_path = str(Path(base_path).parent)

        dialog = BatchFileDialog(
            selection_values=AVAILABLE_OPTIONS_FOR_BATCH,
            directory=base_path,
            parent=self,
        )
        dialog.setWindowTitle("Select Directory")

        if dialog.exec_():
            # Get the values corresponding to the checkboxes that are checked
            options_selected = dialog.get_selected_options()

            # Get the selected directory
            path = dialog.selectedFiles()[0]
            self._batch_process_path = path
            dialog.close()
            dialog.deleteLater()
            return path, options_selected
        else:
            dialog.close()
            dialog.deleteLater()
            return None, None

    def _update_base_name_data(self):
        base_name = self._data_storage_instance.file_name.value
        self.base_name_LineEdit_data.setText(Path(base_name).stem)
        self.base_name_LineEdit_img.setText(Path(base_name).stem)

    def _connect_signals(self):
        self.browse_file_data.clicked.connect(self._browse_file_data)
        self.browse_file_img.clicked.connect(self._browse_file_img)
        self._data_storage_instance.file_name.value_changed.connect(
            self._update_base_name_data
        )

    def setup_ui(self):
        """Load the .ui file and set icons."""
        uic.loadUi(self.UI_FILE, self)  # load QtDesigner .ui file
        self._connect_signals()
        self.browse_file_icon = QIcon(str(ICONS / "folder-open-line.svg"))
        self.browse_file_data.setIcon(self.browse_file_icon)
        self.browse_file_img.setIcon(self.browse_file_icon)
        self.abort_batch_button.setStyleSheet(
            "background-color : #7C0A02; color : white"
        )
        self._hide_abort_batch_button()

    def _hide_abort_batch_button(self):
        self.abort_batch_button.setVisible(False)
        self.abort_batch_button.setEnabled(False)
        self.batch_processing_button.setVisible(True)
        self.batch_processing_button.setEnabled(True)

    def _show_abort_batch_button(self):
        self.abort_batch_button.setVisible(True)
        self.abort_batch_button.setEnabled(True)
        self.batch_processing_button.setVisible(False)
        self.batch_processing_button.setEnabled(False)

    def closeEvent(self, event):
        self.closing.emit()
        event.accept()


class ExportController:
    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        data_storage_instance: DataStorage,
        parent=None,
    ):
        self.viewer = viewer
        self._data_storage_instance = data_storage_instance
        self.widget = _exportwidget(self._data_storage_instance, parent)
        self.abort_timer = QTimer(parent)
        self.abort_timer.timeout.connect(self.abort_timer_timeout)

        self.current_date = self._get_current_date()
        self._connect_callbacks()

    def _get_current_date(self):
        now = datetime.now()
        return now.strftime("%Y%m%d")

    def _export_arcos_data(self):
        if self._data_storage_instance.arcos_output.value.empty:
            show_info("No data to export, run arcos first")
        else:
            path = Path(self.widget.file_LineEdit_data.text())
            output_name = f"{self.current_date}_{self.widget.base_name_LineEdit_data.text()}_arcos_output.csv"
            outpath = os.path.join(path, output_name)
            self._data_storage_instance.arcos_output.value.to_csv(outpath, index=False)
            show_info(f"wrote csv file to {outpath}")

    def _export_arcos_stats(self):
        if self._data_storage_instance.columns.value.object_id is None:
            show_info("No Stats are calculated without a track id column selected")

        elif self._data_storage_instance.arcos_stats.value.empty:
            show_info("No data to export, run arcos first")

        else:
            path = Path(self.widget.file_LineEdit_data.text())
            output_name = f"{self.current_date}_{self.widget.base_name_LineEdit_data.text()}_arcos_stats.csv"
            outpath = os.path.join(path, output_name)
            self._data_storage_instance.arcos_stats.value.to_csv(outpath, index=False)
            show_info(f"wrote csv file to {outpath}")

    def _export_arcos_params(self):
        path = self.widget._browse_parmeters_export()
        if path is None:
            return
        self._data_storage_instance.export_to_yaml(path)
        show_info(f"wrote yaml file to {path}")

    def _import_arcos_params(self):
        path, what_to_import = self.widget._browse_parmeters_import()
        if path is None:
            return
        if not what_to_import:
            show_info("No settings selected to import")
            return
        self._data_storage_instance.import_from_yaml(path, what_to_import)
        show_info(f"imported yaml file from {path}")

    def _export_image_series(self):
        if self.viewer.layers == []:
            show_info("No layers to export")
        else:
            path = Path(self.widget.file_LineEdit_img.text())
            output_name = f"{self.current_date}_{self.widget.base_name_LineEdit_img.text()}_arcos_output"
            outpath = os.path.join(path, output_name)
            MovieExporter(
                self.viewer,
                self.widget.automatic_size_img.isChecked(),
                outpath,
                self.widget.spinBox_width_img.value(),
                self.widget.spinBox_height_img.value(),
            ).run()

    def _connect_callbacks(self):
        self.widget.closing.connect(self.closeEvent)
        self.widget.data_export_button.clicked.connect(self._export_arcos_data)
        self.widget.stats_export_button.clicked.connect(self._export_arcos_stats)
        self.widget.param_export_button.clicked.connect(self._export_arcos_params)
        self.widget.param_import_button.clicked.connect(self._import_arcos_params)
        self.widget.img_seq_export_button.clicked.connect(self._export_image_series)
        self.widget.batch_processing_button.clicked.connect(self.batch_processing)
        self.widget.abort_batch_button.clicked.connect(self._abort_batch)

    def batch_processing(self):
        inpath, what_to_export = self.widget._browse_batch_output()
        if inpath is None:
            return
        self.batch_worker = BatchProcessor(
            inpath,
            self._data_storage_instance.arcos_parameters.value,
            self._data_storage_instance.columns.value,
            self._data_storage_instance.min_max_tracklenght.value[0],
            self._data_storage_instance.min_max_tracklenght.value[1],
            what_to_export,
        )

        self.show_activity_dock(True)
        # Connect signals and slots - assuming batch_worker has finished, aborted, and run methods, and signals
        self.batch_worker.finished.connect(self._on_batch_finish)
        self.batch_worker.errored.connect(self._batch_error)

        self.batch_worker.new_total_files.connect(self.update_progress_files)
        self.batch_worker.new_total_filters.connect(self.update_progress_filters)
        self.batch_worker.finished.connect(self.abort_timer_stop)
        self.batch_worker.aborted.connect(self.abort_timer_stop)

        self.batch_worker.start()
        self.widget._show_abort_batch_button()

    def update_progress_files(self, value):
        if value < 2:
            self.pbar_files = progress(desc="Files")
            return
        self.pbar_files = progress(total=value, desc="Files")
        self.batch_worker.progress_update_files.connect(
            lambda: self.pbar_files.update(1)
        )
        self.batch_worker.aborted.connect(self._close_progress)

    def update_progress_filters(self, value):
        if value < 2:
            return
        self.pbar_filters = progress(
            total=value, nest_under=self.pbar_files, desc="Positions / Filters"
        )
        self.batch_worker.progress_update_filters.connect(
            lambda: self.pbar_filters.update(1)
        )

    def abort_worker(self):
        self.batch_worker.quit()

    def abort_timer_start(self):
        self.abort_timer.start(1000)

    def abort_timer_timeout(self):
        self.update_abort_button()

    def abort_timer_stop(self):
        self.abort_timer.stop()
        self._close_progress()
        self.widget.abort_batch_button.setText("Abort Batch Processing")

    def update_abort_button(self):
        self.dots = self.widget.abort_batch_button.text().count(".")
        if self.dots < 3:
            self.widget.abort_batch_button.setText(
                self.widget.abort_batch_button.text() + "."
            )
        else:
            self.widget.abort_batch_button.setText("Aborting")

    def _on_batch_finish(self):
        show_info("Batch processing finished")
        self.widget._hide_abort_batch_button()
        self._close_progress()

    def _close_progress(self):
        try:
            self.pbar_filters.close()
        except AttributeError:
            pass
        try:
            self.pbar_files.close()
        except AttributeError:
            pass
        self.show_activity_dock(False)

    def _batch_error(self, error_message: Exception):
        show_info(f"{error_message}")
        traceback.print_exception(None, error_message, error_message.__traceback__)

    def show_activity_dock(self, state=True):
        # show/hide activity dock if there is actual progress to see
        try:
            with warnings.catch_warnings():
                # suppress FutureWarning for now: https://github.com/napari/napari/issues/4598
                warnings.simplefilter(action="ignore", category=FutureWarning)
                self.viewer.window._status_bar._toggle_activity_dock(state)
        except AttributeError:
            print("show_activity_dock failed")

    def closeEvent(self):
        self._abort_batch()

    def _abort_batch(self):
        try:
            self.batch_worker.quit()
            self.widget.abort_batch_button.setEnabled(False)
            self.widget.abort_batch_button.setText("Aborting")
            self.abort_timer_start()
        except (AttributeError, RuntimeError):
            pass
        show_info("Aborting Batch Processing")



"""Module for arcos wrapper functions."""

from __future__ import annotations

import os
import warnings
from itertools import product
from pathlib import Path
from typing import Callable

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from arcos4py import ARCOS
from arcos4py.plotting import NoodlePlot, statsPlots
from arcos4py.tools import (
    calculate_statistics,
    calculate_statistics_per_frame,
    estimate_eps,
    filterCollev,
)
from arcos4py.tools._detect_events import DataFrameTracker, Linker
from arcos_gui.processing._data_storage import ArcosParameters, columnnames
from arcos_gui.processing._preprocessing_utils import (
    calculate_measurement,
    check_for_collid_column,
    create_file_names,
    create_output_folders,
    filter_data,
)
from arcos_gui.tools import AVAILABLE_OPTIONS_FOR_BATCH, OPERATOR_DICTIONARY
from napari.qt.threading import WorkerBase, WorkerBaseSignals
from qtpy.QtCore import Signal


class customARCOS(ARCOS):
    """Custom ARCOS class with replaced trackCollev method.

    The trackCollev method is replaced with a custom version that emits a signal
    for progress updates. This signal is connected to the ArcosWidget, which
    updates the progress bar. The custom trackCollev method also checks for the
    abort_requested flag and aborts the tracking if it is set to True.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._abort_requested = False
        self.progress_update = None

    def trackCollev(
        self,
        eps: float = 1,
        epsPrev: float | None = None,
        minClsz: int = 1,
        nPrev: int = 1,
        clusteringMethod: str = "dbscan",
        linkingMethod: str = "nearest",
        minSamples: int | None = None,
    ) -> pd.DataFrame:
        linker = Linker(
            eps=eps,
            epsPrev=epsPrev,
            minClSz=minClsz,
            minSamples=minSamples,
            clusteringMethod=clusteringMethod,
            linkingMethod=linkingMethod,
            nPrev=nPrev,
            predictor=False,
            nJobs=1,
        )
        tracker = DataFrameTracker(
            linker=linker,
            coordinates_column=self.posCols,
            frame_column=self.frame_column,
            id_column=self.id_column,
            bin_meas_column=self.bin_col,
            collid_column=self.clid_column,
        )
        df_list = []

        total = self.data[self.frame_column].nunique()

        if self.progress_update:
            self.progress_update.emit("total", total)

        for timepoint in tracker.track(self.data):
            if self.abort_requested:
                self._abort_requested = False
                return pd.DataFrame()
            df_list.append(timepoint)

            if self.progress_update:
                self.progress_update.emit("update", 1)

        if self.progress_update:
            self.progress_update.emit("reset", 0)

        df_out = pd.concat(df_list, axis=0)

        return df_out.query(f"{self.clid_column} != -1").reset_index(drop=True)

    def quit(self):
        self._abort_requested = True

    @property
    def abort_requested(self):
        return self._abort_requested


def empty_std_out(*args, **kwargs):
    pass


def init_arcos_object(
    df_in: pd.DataFrame,
    posCols: list,
    measurement_name: str,
    frame_col_name: str,
    track_id_col_name: str,
    progress_update_signal: Signal | None = None,
):
    """
    Initialize arcos object from pandas dataframe

    Parameters
    ----------
    df_in : pd.DataFrame
        input dataframe
    posCols : list
        list of position columns
    measurement_name : str
        name of measurement column
    frame_col_name : str
        name of frame column
    track_id_col_name : str
        name of track id column
    std_out_func : Callable
        function to print to console or gui
    """

    collid_name = "collid"

    df_in = check_for_collid_column(df_in, collid_name)

    # checks if this part of the function has to be run,
    # depends on the parameters changed in arcos widget

    # create arcos object, run arcos
    arcos = customARCOS(
        data=df_in,
        posCols=posCols,
        frame_column=frame_col_name,
        id_column=track_id_col_name,
        measurement_column=measurement_name,
        clid_column=collid_name,
    )
    arcos.progress_update = progress_update_signal
    return arcos


def binarization(
    arcos: customARCOS,
    interpolate_meas: bool,
    clip_meas: bool,
    clip_low: float,
    clip_high: float,
    smooth_k: int,
    bias_k: int,
    bin_peak_threshold: float,
    bin_threshold: float,
    polyDeg: int,
    bias_method: str,
) -> customARCOS:
    """
    Binarize measurement data

    Parameters
    ----------
    arcos : customARCOS
        arcos object
    interpolate_meas : bool
        interpolate measurement data
    clip_meas : bool
        clip measurement data
    clip_low : float
        lower clip value
    clip_high : float
        higher clip value
    smooth_k : int
        smoothing kernel size
    bias_k : int
        bias kernel size
    bin_peak_threshold : float
        peak threshold
    bin_threshold : float
        binarization threshold
    polyDeg : int
        polynomial degree
    bias_method : str
        bias method

    Returns
    -------
    arcos_bin : customARCOS
        arcos object with binarized measurement data
    """
    # if corresponding checkbox was selected run interpolate measurments
    if interpolate_meas:
        arcos.interpolate_measurements()

    # if corresponding checkbock was selected run clip_measuremnts
    if clip_meas:
        arcos.clip_meas(
            clip_low=clip_low,
            clip_high=clip_high,
        )

    # binarize data and update ts variable
    # update from where to run
    arcos.bin_measurements(
        smoothK=smooth_k,
        biasK=bias_k,
        peakThr=bin_peak_threshold,
        binThr=bin_threshold,
        polyDeg=polyDeg,
        biasMet=bias_method,
    )

    return arcos


def detect_events(
    arcos: customARCOS,
    neighbourhood_size: float,
    epsPrev: float | None,
    min_clustersize: int,
    nPrev_value: int,
):
    """
    Detect collective events with arcos trackCollev method.

    Parameters
    ----------
    arcos : customARCOS
        arcos object
    measbin_col : str
        name of binarized measurement column
    neighbourhood_size : float
        neighbourhood size to consider for event detection
    min_clustersize : int
        minimum cluster size to consider for event detection
    nPrev_value : int
        number of previous frames to consider for event detection

    Returns
    -------
    arcos_events : pd.DataFrame
        dataframe with detected events
    """
    _bin_col = arcos.bin_col
    if 1 not in arcos.data[_bin_col].values:
        return None

    arcos_events = arcos.trackCollev(
        eps=neighbourhood_size,
        epsPrev=epsPrev,
        minClsz=min_clustersize,
        nPrev=nPrev_value,
    )
    return arcos_events


def get_eps(arcos: customARCOS, method: str, minClustersize: int, current_eps: float):
    """
    Estimate eps value for arcos trackCollev method.

    Parameters
    ----------
    arcos : customARCOS
        arcos object
    method : str
        method to estimate eps value
    minClustersize : int
        minimum cluster size to consider for event detection
    current_eps : float | None
        current eps value, will be returned if method is manual

    Returns
    -------
    eps : float
        eps value
    """
    methods = ["manual", "kneepoint", "mean"]
    if method not in methods:
        raise ValueError(f"Method must be one of {methods}")

    if method == "kneepoint":
        eps = estimate_eps(
            data=arcos.data[arcos.data[arcos.bin_col] == 1],
            method="kneepoint",
            pos_cols=arcos.posCols,
            frame_col=arcos.frame_column,
            n_neighbors=minClustersize,
            plot=False,
        )
        return round(eps, 2)

    if method == "mean":
        eps = estimate_eps(
            arcos.data,
            method="mean",
            pos_cols=arcos.posCols,
            frame_col=arcos.frame_column,
            n_neighbors=minClustersize,
            plot=False,
        )
        return round(eps, 2)
    return round(current_eps, 2)


def filtering_arcos_events(
    detected_events_df: pd.DataFrame,
    frame_col_name: str,
    collid_name: str,
    track_id_col_name: str,
    min_dur: int,
    total_event_size: int,
):
    """
    Filter detected events with arcos filterCollev method.

    Parameters
    ----------
    detected_events_df : pd.DataFrame
        dataframe with detected events
    frame_col_name : str
        name of frame column
    collid_name : str
        name of collid column
    track_id_col_name : str
        name of track id column
    min_dur : int
        minimum duration of events
    total_event_size : int
        minimum size of events

    Returns
    -------
    filtered_events_df : pd.DataFrame
        dataframe with filtered events
    """
    if track_id_col_name:
        filterer = filterCollev(
            data=detected_events_df,
            frame_column=frame_col_name,
            collid_column=collid_name,
            obj_id_column=track_id_col_name,
        )
        arcos_filtered = filterer.filter(
            coll_duration=min_dur,
            coll_total_size=total_event_size,
        )
    else:
        # filter dataframe by duraiton of events
        detect_events_df = detected_events_df.copy()
        detect_events_df["duration"] = detect_events_df.groupby(
            [frame_col_name, collid_name]
        )[frame_col_name].transform("count")
        arcos_filtered = detect_events_df[detect_events_df["duration"] >= min_dur]
        arcos_filtered = arcos_filtered.drop(columns=["duration"])

    # makes filtered collids sequential
    clid_np = arcos_filtered[collid_name].to_numpy()
    clids_sorted_i = np.argsort(clid_np)
    clids_reverse_i = np.argsort(clids_sorted_i)
    clid_np_sorted = clid_np[(clids_sorted_i)]
    grouped_array_clids = np.split(
        clid_np_sorted,
        np.unique(clid_np_sorted, axis=0, return_index=True)[1][1:],
    )
    seq_colids = np.concatenate(
        [np.repeat(i, value.shape[0]) for i, value in enumerate(grouped_array_clids)],
        axis=0,
    )[clids_reverse_i]
    seq_colids_from_one = np.add(seq_colids, 1)
    arcos_filtered = arcos_filtered.copy()
    arcos_filtered.loc[:, collid_name] = seq_colids_from_one

    return arcos_filtered


def calculate_arcos_stats(
    df_arcos_filtered: pd.DataFrame,
    frame_col: str,
    collid_name: str,
    object_id_name: str,
    posCols: list,
):
    """Wrapper for calcCollevStats().

    Parameters
    ----------
    df_arcos_filtered : pd.DataFrame
        dataframe with filtered events
    frame_col : str
        name of frame column
    collid_name : str
        name of collid column
    object_id_col_name : str
        name of object id column
    posCols : list
        list of position columns

    Returns
    -------
    df_arcos_stats : pd.DataFrame
        dataframe with statistics for each event
    """
    df_arcos_stats = calculate_statistics(
        df_arcos_filtered, frame_col, collid_name, object_id_name, posCols
    )
    return df_arcos_stats


class arcos_worker_base_signals(WorkerBaseSignals):
    binarization_finished = Signal(tuple)
    tracking_finished = Signal()
    new_arcos_output = Signal(tuple)
    new_eps = Signal(float)
    started = Signal()
    finished = Signal()
    aborted = Signal(object)
    arcos_progress_update = Signal(str, int)


class arcos_worker(WorkerBase):
    """Runs arcos with the current parameters defined in the ArcosWidget.

    Updates the data storage with the results. what_to_run is a set of strings
    indicating what to run. The strings correspond to specific steps in the
    arcos pipeline. The steps are:
        - 'binarization': initializes a new customARCOS object and runs the binarization.
        - 'tracking': runs the event detection.
        - 'filtering': runs the event filtering.
    """

    def __init__(
        self,
        what_to_run: set,
        std_out: Callable,
        arcos_parameters: ArcosParameters = ArcosParameters(),
        columns: columnnames = columnnames(),
        filtered_data: pd.DataFrame = pd.DataFrame(),
        arcos_object: customARCOS | None = None,
        arcos_raw_output: pd.DataFrame | None = None,
    ):
        """Constructor.

        Parameters
        ----------
        what_to_run : set
            set of strings indicating what to run
        std_out : Callable
            function to print to the console
        wait_for_parameter_update : bool, optional
            if True, the worker will wait for the parameters to be updated before running
        """
        super().__init__(SignalsClass=arcos_worker_base_signals)
        self.what_to_run = what_to_run
        self.std_out = std_out
        self.arcos_parameters: ArcosParameters = arcos_parameters
        self.columns: columnnames = columns
        self.filtered_data: pd.DataFrame = filtered_data
        if arcos_object is None:
            arcos_object = init_arcos_object(
                df_in=pd.DataFrame(columns=["x", "t", "m", "id"]),
                posCols=["x"],
                frame_col_name="t",
                track_id_col_name="id",
                measurement_name="m",
                progress_update_signal=self.arcos_progress_update,
            )
        else:
            # attatch progress signal to arcos object if the arcos object is not None
            arcos_object.progress_update = self.arcos_progress_update

        self.arcos_object: customARCOS = arcos_object

        if arcos_raw_output is None:
            arcos_raw_output = pd.DataFrame(
                columns=["t", "id", "collid", "x", "y", "m"], data=[]
            )
        self.arcos_raw_output: pd.DataFrame = arcos_raw_output

    def quit(self) -> None:
        """Quit the worker. Sets the abort_requested flag to True.
        Reimplemented so that it also sets the arcos_abort_requested flag to True.
        """
        self._abort_requested = True
        self.arcos_object.quit()
        super().quit()

    def run_binarization(self):
        if self.filtered_data.empty:
            self.std_out("No data loaded. Load first using the import data tab.")
            return
        self.started.emit()

        if self.abort_requested:
            return
        self.arcos_object = init_arcos_object(
            df_in=self.filtered_data,
            posCols=self.columns.posCol,
            measurement_name=self.columns.measurement_column,
            frame_col_name=self.columns.frame_column,
            track_id_col_name=self.columns.object_id,
            progress_update_signal=self.arcos_progress_update,
        )
        if self.abort_requested:
            return

        self.arcos_object = binarization(
            arcos=self.arcos_object,
            interpolate_meas=self.arcos_parameters.interpolate_meas.value,
            clip_meas=self.arcos_parameters.clip_meas.value,
            clip_low=self.arcos_parameters.clip_low.value,
            clip_high=self.arcos_parameters.clip_high.value,
            smooth_k=self.arcos_parameters.smooth_k.value,
            bias_k=self.arcos_parameters.bias_k.value,
            polyDeg=self.arcos_parameters.polyDeg.value,
            bin_threshold=self.arcos_parameters.bin_threshold.value,
            bin_peak_threshold=self.arcos_parameters.bin_peak_threshold.value,
            bias_method=self.arcos_parameters.bias_method.value,
        )
        if self.abort_requested:
            return
        self.binarization_finished.emit(
            (
                self.arcos_object.bin_col,
                self.arcos_object.resc_col,
                self.arcos_object.data,
            )
        )
        self.columns.measurement_bin = self.arcos_object.bin_col
        self.columns.measurement_resc = self.arcos_object.resc_col
        self.what_to_run.remove("binarization")

    def run_tracking(self):
        try:
            bin_col = self.columns.measurement_bin
            n_bin = self.arcos_object.data[bin_col].nunique()

        except KeyError:
            n_bin = 0
        if n_bin < 2:
            self.std_out("No Binarized Data. Adjust Binazation Parameters.")
            return

        if self.abort_requested:
            return

        eps = get_eps(
            arcos=self.arcos_object,
            method=self.arcos_parameters.eps_method.value,
            minClustersize=self.arcos_parameters.min_clustersize.value,
            current_eps=self.arcos_parameters.neighbourhood_size.value,
        )

        if self.abort_requested:
            return

        self.new_eps.emit(eps)

        self.arcos_raw_output = detect_events(
            arcos=self.arcos_object,
            neighbourhood_size=eps,
            epsPrev=self.arcos_parameters.epsPrev.value,
            min_clustersize=self.arcos_parameters.min_clustersize.value,
            nPrev_value=self.arcos_parameters.nprev.value,
        )
        if self.abort_requested:
            return
        self.tracking_finished.emit()
        self.what_to_run.remove("tracking")

    def run_filtering(self):
        if self.arcos_raw_output.empty:
            self.std_out(
                "No Collective Events detected. Adjust Event Detection Parameters."
            )
            return

        collid_name = "collid"
        if self.abort_requested:
            return
        arcos_df_filtered = filtering_arcos_events(
            detected_events_df=self.arcos_raw_output,
            frame_col_name=self.columns.frame_column,
            collid_name=collid_name,
            track_id_col_name=self.columns.object_id,
            min_dur=self.arcos_parameters.min_dur.value,
            total_event_size=self.arcos_parameters.total_event_size.value,
        )
        if arcos_df_filtered.empty:
            self.std_out("No Collective Events detected.Adjust Filtering parameters.")
            return
        if self.abort_requested:
            return
        arcos_stats = calculate_arcos_stats(
            df_arcos_filtered=arcos_df_filtered,
            frame_col=self.columns.frame_column,
            collid_name=collid_name,
            object_id_name=self.columns.object_id,
            posCols=self.columns.posCol,
        )
        arcos_stats = arcos_stats.dropna()
        if self.abort_requested:
            return
        self.new_arcos_output.emit((arcos_df_filtered, arcos_stats))
        self.what_to_run.clear()

    def run(
        self,
    ):
        """Run arcos with input parameters.

        Runs only or only from as far as specified in the what_to_run set.
        """
        self.started.emit()

        try:
            if "binarization" in self.what_to_run and not self.abort_requested:
                self.run_binarization()

            if "tracking" in self.what_to_run and not self.abort_requested:
                self.run_tracking()

            if "filtering" in self.what_to_run and not self.abort_requested:
                self.run_filtering()

        except Exception as e:
            self.errored.emit(e)

        finally:
            self.finished.emit()


class TemporaryMatplotlibBackend:
    def __init__(self, backend="Agg"):
        self.temp_backend = backend
        self.original_backend = matplotlib.get_backend()

    def __enter__(self):
        plt.switch_backend(self.temp_backend)

    def __exit__(self, *args):
        plt.switch_backend(self.original_backend)


class BatchProcessorSignals(WorkerBaseSignals):
    finished = Signal()
    progress_update_files = Signal()
    progress_update_filters = Signal()
    new_total_files = Signal(int)
    new_total_filters = Signal(int)
    aborted = Signal()


class BatchProcessor(WorkerBase):
    """Runs Arcos in batch mode with the current parameters defined in the ArcosWidget."""

    def __init__(
        self,
        input_path: str,
        arcos_parameters: ArcosParameters,
        columnames: columnnames,
        min_tracklength: int,
        max_tracklength: int,
        what_to_export: list[str],
    ):
        super().__init__(SignalsClass=BatchProcessorSignals)
        self.input_path = input_path
        self.arcos_parameters = arcos_parameters
        self.columnames = columnames
        self.min_track_length = min_tracklength
        self.max_track_length = max_tracklength
        self.what_to_export = what_to_export

    def _create_fileendings_list(self):
        """Create a list o file endings for the files to be exported."""
        fileendings = []
        corresponding_fileendings = [".csv", ".csv", ".csv", ".svg", ".svg"]
        for option in self.what_to_export:
            if option in AVAILABLE_OPTIONS_FOR_BATCH:
                # get index of option in AVAILABLE_OPTIONS_FOR_BATCH
                index = AVAILABLE_OPTIONS_FOR_BATCH.index(option)
                # get corresponding file ending
                file_ending = corresponding_fileendings[index]
                # add file ending to list
                fileendings.append(file_ending)
        return fileendings

    def run_arcos_batch(self, df):
        """Run arcos with input parameters.

        Runs only or only from as far as specified in the what_to_run set.
        """
        arcos = init_arcos_object(
            df,
            self.columnames.posCol,
            self.columnames.measurement_column,
            self.columnames.frame_column,
            self.columnames.object_id,
        )
        arcos = binarization(
            arcos=arcos,
            interpolate_meas=self.arcos_parameters.interpolate_meas.value,
            clip_meas=self.arcos_parameters.clip_meas.value,
            clip_low=self.arcos_parameters.clip_low.value,
            clip_high=self.arcos_parameters.clip_high.value,
            smooth_k=self.arcos_parameters.smooth_k.value,
            bias_k=self.arcos_parameters.bias_k.value,
            polyDeg=self.arcos_parameters.polyDeg.value,
            bin_threshold=self.arcos_parameters.bin_threshold.value,
            bin_peak_threshold=self.arcos_parameters.bin_peak_threshold.value,
            bias_method=self.arcos_parameters.bias_method.value,
        )
        eps = get_eps(
            arcos=arcos,
            method=self.arcos_parameters.eps_method.value,
            minClustersize=self.arcos_parameters.min_clustersize.value,
            current_eps=self.arcos_parameters.neighbourhood_size.value,
        )
        arcos_raw_output = detect_events(
            arcos=arcos,
            neighbourhood_size=eps,
            epsPrev=self.arcos_parameters.epsPrev.value,
            min_clustersize=self.arcos_parameters.min_clustersize.value,
            nPrev_value=self.arcos_parameters.nprev.value,
        )
        arcos_df_filtered = filtering_arcos_events(
            detected_events_df=arcos_raw_output,
            frame_col_name=self.columnames.frame_column,
            collid_name="collid",
            track_id_col_name=self.columnames.object_id,
            min_dur=self.arcos_parameters.min_dur.value,
            total_event_size=self.arcos_parameters.total_event_size.value,
        )
        if arcos_df_filtered.empty:
            return arcos_df_filtered, pd.DataFrame()

        arcos_stats = calculate_arcos_stats(
            df_arcos_filtered=arcos_df_filtered,
            frame_col=self.columnames.frame_column,
            collid_name="collid",
            object_id_name=self.columnames.object_id,
            posCols=self.columnames.posCol,
        )
        arcos_stats = arcos_stats.dropna()

        return arcos_df_filtered, arcos_stats

    def run(self):
        """Run arcos with input parameters.

        Runs only or only from as far as specified in the what_to_run set.
        """
        self.started.emit()
        try:
            summary_stats = summary_stats = {
                key: []
                for key in [
                    "file",
                    "fov",
                    "additional_filter",
                    "event_count",
                    "avg_total_size",
                    "avg_total_size_std",
                    "avg_duration",
                    "avg_duration_std",
                ]
            }  # noqa E501
            with TemporaryMatplotlibBackend("Agg"):
                # check that what_to_export
                if not self.what_to_export:
                    self.errored.emit(ValueError("No export selected"))
                    return

                print(f"Exporting {self.what_to_export}")

                # check that there are only valid export options
                valid_export_options = AVAILABLE_OPTIONS_FOR_BATCH
                for option in self.what_to_export:
                    if option not in valid_export_options:
                        self.errored.emit(ValueError(f"Invalid export option {option}"))
                        return

                # remove statsplot and noodleplot from what_to_export if no object id is given
                if self.columnames.object_id is None:
                    self.what_to_export = [
                        option
                        for option in self.what_to_export
                        if option not in ["statsplot", "noodleplot"]
                    ]
                    print("No object id given. Skipping statsplot and noodleplot.")

                file_list = [
                    os.path.join(self.input_path, file)
                    for file in os.listdir(self.input_path)
                    if file.endswith(".csv") or file.endswith(".csv.gz")
                ]
                self.new_total_files.emit(len(file_list))

                base_path, _ = create_output_folders(
                    self.input_path, self.what_to_export
                )
                for file in file_list:
                    if self.abort_requested:
                        self.aborted.emit()
                        break

                    pth = Path(file)

                    file_name = pth.with_suffix("").stem
                    print(f"Processing file {file_name}")
                    df = pd.read_csv(file, engine="pyarrow")

                    meas_col, df = calculate_measurement(
                        data=df,
                        operation=self.columnames.measurement_math_operation,
                        in_meas_1_name=self.columnames.measurement_column_1,
                        in_meas_2_name=self.columnames.measurement_column_2,
                        op_dict=OPERATOR_DICTIONARY,
                    )
                    self.columnames.measurement_column = meas_col

                    if self.columnames.position_id is not None:
                        position_ids = df[self.columnames.position_id].unique()
                    else:
                        position_ids = [None]

                    if self.columnames.additional_filter_column is not None:
                        additional_filters = df[
                            self.columnames.additional_filter_column
                        ].unique()
                    else:
                        additional_filters = [None]

                    iterator_fov_filter = list(
                        product(position_ids, additional_filters)
                    )
                    self.new_total_filters.emit(len(iterator_fov_filter))

                    for fov, additional_filter in iterator_fov_filter:
                        if self.abort_requested:
                            self.aborted.emit()
                            break

                        # add new row to summary stats
                        for key in summary_stats.keys():
                            summary_stats[key].append(pd.NA)

                        # update general stats that should be present for all iterations
                        summary_stats["file"][-1] = file_name
                        summary_stats["fov"][-1] = fov if fov is not None else pd.NA
                        summary_stats["additional_filter"][-1] = (
                            additional_filter
                            if additional_filter is not None
                            else pd.NA
                        )

                        df_filtered = filter_data(
                            df_in=df,
                            field_of_view_id_name=self.columnames.position_id,
                            frame_name=self.columnames.frame_column,
                            track_id_name=self.columnames.object_id,
                            measurement_name=self.columnames.measurement_column,
                            additional_filter_column_name=self.columnames.additional_filter_column,
                            posCols=self.columnames.posCol,
                            fov_val=fov,
                            additional_filter_value=additional_filter,
                            min_tracklength_value=self.min_track_length,
                            max_tracklength_value=self.max_track_length,
                            frame_interval=1,
                            st_out=empty_std_out,
                        )[0]
                        if df_filtered.empty:
                            # set event count to 0, rest is already set to nan
                            summary_stats["event_count"][-1] = 0

                            position_id_str = (
                                f"{self.columnames.position_id}:{fov}"
                                if self.columnames.position_id is not None
                                and fov is not None
                                else ""
                            )
                            additional_filter_str = (
                                f"{self.columnames.additional_filter_column}:{additional_filter}"
                                if self.columnames.additional_filter_column is not None
                                and additional_filter is not None
                                else ""
                            )
                            connector = (
                                " and "
                                if position_id_str and additional_filter_str
                                else ""
                            )
                            for_str = (
                                "for "
                                if position_id_str or additional_filter_str
                                else ""
                            )
                            error_message = f"No data for file {file} {for_str}{position_id_str}{connector}{additional_filter_str}"  # noqa E501
                            self.progress_update_filters.emit()
                            print(error_message)
                            continue

                        posx = self.columnames.posCol[0]
                        posy = self.columnames.posCol[1]
                        if len(self.columnames.posCol) == 2:
                            posz = None
                        else:
                            posz = self.columnames.posCol[2]

                        arcos_df_filtered, arcos_stats = self.run_arcos_batch(
                            df_filtered
                        )

                        if arcos_df_filtered.empty:
                            # set event count to 0, rest is already set to nan
                            summary_stats["event_count"][-1] = 0

                            print(
                                f"No events detected for file {file} filters fov:{fov} additional:{additional_filter}"
                            )  # noqa E501
                            self.progress_update_filters.emit()
                            continue

                        # update summary stats
                        summary_stats["event_count"][-1] = arcos_stats[
                            "collid"
                        ].nunique()
                        summary_stats["avg_total_size"][-1] = arcos_stats[
                            "total_size"
                        ].mean()
                        summary_stats["avg_total_size_std"][-1] = arcos_stats[
                            "total_size"
                        ].std()
                        summary_stats["avg_duration"][-1] = arcos_stats[
                            "duration"
                        ].mean()
                        summary_stats["avg_duration_std"][-1] = arcos_stats[
                            "duration"
                        ].std()

                        out_file_name = create_file_names(
                            base_path,
                            file_name,
                            self.what_to_export,
                            self._create_fileendings_list(),
                            fov,
                            additional_filter,
                            self.columnames.position_id,
                            self.columnames.additional_filter_column,
                        )
                        if "arcos_output" in self.what_to_export:
                            arcos_df_filtered.to_csv(
                                out_file_name["arcos_output"],
                                index=False,
                            )
                        if "arcos_stats" in self.what_to_export:
                            arcos_stats.to_csv(
                                out_file_name["arcos_stats"],
                                index=False,
                            )
                        if "per_frame_statistics" in self.what_to_export:
                            arcos_stats_per_frame = calculate_statistics_per_frame(
                                data=arcos_df_filtered,
                                frame_column=self.columnames.frame_column,
                                collid_column="collid",
                                pos_columns=self.columnames.posCol,
                            )
                            arcos_stats_per_frame.to_csv(
                                out_file_name["per_frame_statistics"],
                                index=False,
                            )

                        if "statsplot" in self.what_to_export:
                            # seaborn future warning is annoying
                            with warnings.catch_warnings():
                                warnings.simplefilter(
                                    action="ignore", category=FutureWarning
                                )
                                arcos_stats_plot = statsPlots(arcos_stats)
                                arcos_stats_plot.plot_events_duration(
                                    "total_size", "duration"
                                )
                                plt.savefig(out_file_name["statsplot"])
                                plt.close()

                        if "noodleplot" in self.what_to_export:
                            noodle_plot = NoodlePlot(
                                df=arcos_df_filtered,
                                colev=self.columnames.collid_name,
                                trackid=self.columnames.object_id,
                                frame=self.columnames.frame_column,
                                posx=posx,
                                posy=posy,
                                posz=posz,
                            )

                            noodle_plot.plot(posx)
                            plt.savefig(out_file_name["noodleplot"])
                            plt.close()

                        self.progress_update_filters.emit()

                    self.progress_update_files.emit()

            summary_stats_df = pd.DataFrame(summary_stats).round(4)
            # drop rows with all nan
            summary_stats_df = summary_stats_df.dropna(how="all", axis=1)
            summary_stats_df.to_csv(
                os.path.join(base_path, "per_file_summary.csv"),
                index=False,
                na_rep="NA",
            )

        except Exception as e:
            self.errored.emit(e)
        finally:
            self.finished.emit()
